# -*- encoding: utf-8 -*-
from __future__ import annotations

from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import warnings

import dask.distributed
import joblib
import numpy as np
import pandas as pd
from ConfigSpace.configuration_space import Configuration, ConfigurationSpace
from scipy.sparse import spmatrix
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.ensemble import VotingClassifier, VotingRegressor
from sklearn.utils.multiclass import type_of_target
from smac.runhistory.runhistory import RunInfo, RunValue
from typing_extensions import Literal

from autosklearn.automl import AutoML, AutoMLClassifier, AutoMLRegressor
from autosklearn.data.validation import (
    SUPPORTED_FEAT_TYPES,
    SUPPORTED_TARGET_TYPES,
    convert_if_sparse,
)
from autosklearn.ensembles.abstract_ensemble import AbstractEnsemble
from autosklearn.ensembles.ensemble_selection import EnsembleSelection
from autosklearn.ensembles.multiobjective_dummy_ensemble import (
    MultiObjectiveDummyEnsemble,
)
from autosklearn.metrics import Scorer
from autosklearn.pipeline.base import BasePipeline
from autosklearn.util.smac_wrap import SMACCallback


class AutoSklearnEstimator(BaseEstimator):
    def __init__(
        self,
        time_left_for_this_task=3600,
        per_run_time_limit=None,
        initial_configurations_via_metalearning=25,
        ensemble_size: int | None = None,
        ensemble_class: Type[AbstractEnsemble] | Literal["default"] | None = "default",
        ensemble_kwargs: Dict[str, Any] | None = None,
        ensemble_nbest=50,
        max_models_on_disc=50,
        seed=1,
        memory_limit=3072,
        include: Optional[Dict[str, List[str]]] = None,
        exclude: Optional[Dict[str, List[str]]] = None,
        resampling_strategy="holdout",
        resampling_strategy_arguments=None,
        tmp_folder=None,
        delete_tmp_folder_after_terminate=True,
        n_jobs: Optional[int] = None,
        dask_client: Optional[dask.distributed.Client] = None,
        disable_evaluator_output=False,
        get_smac_object_callback=None,
        smac_scenario_args=None,
        logging_config=None,
        metadata_directory=None,
        metric: Scorer | Sequence[Scorer] | None = None,
        scoring_functions: Optional[List[Scorer]] = None,
        load_models: bool = True,
        get_trials_callback: SMACCallback | None = None,
        dataset_compression: Union[bool, Mapping[str, Any]] = True,
        allow_string_features: bool = True,
    ):
        """
        Parameters
        ----------
        time_left_for_this_task : int, optional (default=3600)
            Time limit in seconds for the search of appropriate
            models. By increasing this value, *auto-sklearn* has a higher
            chance of finding better models.

        per_run_time_limit : int, optional (default=1/10 of time_left_for_this_task)
            Time limit for a single call to the machine learning model.
            Model fitting will be terminated if the machine learning
            algorithm runs over the time limit. Set this value high enough so
            that typical machine learning algorithms can be fit on the
            training data.

        initial_configurations_via_metalearning : int, optional (default=25)
            Initialize the hyperparameter optimization algorithm with this
            many configurations which worked well on previously seen
            datasets. Disable if the hyperparameter optimization algorithm
            should start from scratch.

        ensemble_size : int, optional
            Number of models added to the ensemble built by *Ensemble
            selection from libraries of models*. Models are drawn with
            replacement. If set to ``0`` no ensemble is fit.

            Deprecated - will be removed in Auto-sklearn 0.16. Please pass
            this argument via ``ensemble_kwargs={"ensemble_size": int}``
            if you want to change the ensemble size for ensemble selection.

        ensemble_class : Type[AbstractEnsemble] | "default", optional (default="default")
            Class implementing the post-hoc ensemble algorithm. Set to
            ``None`` to disable ensemble building or use :class:`SingleBest`
            to obtain only use the single best model instead of an
            ensemble.

            If set to "default" it will use :class:`EnsembleSelection` for
            single-objective problems and :class:`MultiObjectiveDummyEnsemble`
            for multi-objective problems.

        ensemble_kwargs : Dict, optional
            Keyword arguments that are passed to the ensemble class upon
            initialization.

        ensemble_nbest : int, optional (default=50)
            Only consider the ``ensemble_nbest`` models when building an
            ensemble. This is inspired by a concept called library pruning
            introduced in `Getting Most out of Ensemble Selection`. This
            is independent of the ``ensemble_class`` argument and this
            pruning step is done prior to constructing an ensemble.

        max_models_on_disc: int, optional (default=50),
            Defines the maximum number of models that are kept in the disc.
            The additional number of models are permanently deleted. Due to the
            nature of this variable, it sets the upper limit on how many models
            can be used for an ensemble.
            It must be an integer greater or equal than 1.
            If set to None, all models are kept on the disc.

        seed : int, optional (default=1)
            Used to seed SMAC. Will determine the output file names.

        memory_limit : int, optional (3072)
            Memory limit in MB for the machine learning algorithm.
            `auto-sklearn` will stop fitting the machine learning algorithm if
            it tries to allocate more than ``memory_limit`` MB.

            **Important notes:**

            * If ``None`` is provided, no memory limit is set.
            * In case of multi-processing, ``memory_limit`` will be *per job*, so the total usage is
              ``n_jobs x memory_limit``.
            * The memory limit also applies to the ensemble creation process.

        include : Optional[Dict[str, List[str]]] = None
            If None, all possible algorithms are used.

            Otherwise, specifies a step and the components that are included in search.
            See ``/pipeline/components/<step>/*`` for available components.

            Incompatible with parameter ``exclude``.

            **Possible Steps**:

            * ``"data_preprocessor"``
            * ``"balancing"``
            * ``"feature_preprocessor"``
            * ``"classifier"`` - Only for when when using ``AutoSklearnClasssifier``
            * ``"regressor"`` - Only for when when using ``AutoSklearnRegressor``

            **Example**:

            .. code-block:: python

                include = {
                    'classifier': ["random_forest"],
                    'feature_preprocessor': ["no_preprocessing"]
                }

        exclude : Optional[Dict[str, List[str]]] = None
            If None, all possible algorithms are used.

            Otherwise, specifies a step and the components that are excluded from search.
            See ``/pipeline/components/<step>/*`` for available components.

            Incompatible with parameter ``include``.

            **Possible Steps**:

            * ``"data_preprocessor"``
            * ``"balancing"``
            * ``"feature_preprocessor"``
            * ``"classifier"`` - Only for when when using ``AutoSklearnClasssifier``
            * ``"regressor"`` - Only for when when using ``AutoSklearnRegressor``

            **Example**:

            .. code-block:: python

                exclude = {
                    'classifier': ["random_forest"],
                    'feature_preprocessor': ["no_preprocessing"]
                }

        resampling_strategy : str | BaseCrossValidator | _RepeatedSplits | BaseShuffleSplit = "holdout"
            How to to handle overfitting, might need to use ``resampling_strategy_arguments``
            if using ``"cv"`` based method or a Splitter object.

            * **Options**
                *   ``"holdout"`` - Use a 67:33 (train:test) split
                *   ``"cv"``: perform cross validation, requires "folds" in ``resampling_strategy_arguments``
                *   ``"holdout-iterative-fit"`` - Same as "holdout" but iterative fit where possible
                *   ``"cv-iterative-fit"``: Same as "cv" but iterative fit where possible
                *   ``"partial-cv"``: Same as "cv" but uses intensification.
                *   ``BaseCrossValidator`` - any BaseCrossValidator subclass (found in scikit-learn model_selection module)
                *   ``_RepeatedSplits`` - any _RepeatedSplits subclass (found in scikit-learn model_selection module)
                *   ``BaseShuffleSplit`` - any BaseShuffleSplit subclass (found in scikit-learn model_selection module)

            If using a Splitter object that relies on the dataset retaining it's current
            size and order, you will need to look at the ``dataset_compression`` argument
            and ensure that ``"subsample"`` is not included in the applied compression
            ``"methods"`` or disable it entirely with ``False``.

        resampling_strategy_arguments : Optional[Dict] = None
            Additional arguments for ``resampling_strategy``, this is required if
            using a ``cv`` based strategy. The default arguments if left as ``None``
            are:

            .. code-block:: python

                {
                    "train_size": 0.67,     # The size of the training set
                    "shuffle": True,        # Whether to shuffle before splitting data
                    "folds": 5              # Used in 'cv' based resampling strategies
                }

            If using a custom splitter class, which takes ``n_splits`` such as
            `PredefinedSplit <https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.KFold.html#sklearn-model-selection-kfold>`_,
            the value of ``"folds"`` will be used.

        tmp_folder : string, optional (None)
            folder to store configuration output and log files, if ``None``
            automatically use ``/tmp/autosklearn_tmp_$pid_$random_number``

        delete_tmp_folder_after_terminate: bool, optional (True)
            remove tmp_folder, when finished. If tmp_folder is None
            tmp_dir will always be deleted

        n_jobs : int, optional, experimental
            The number of jobs to run in parallel for ``fit()``. ``-1`` means
            using all processors.

            **Important notes**:

            * By default, Auto-sklearn uses one core.
            * Ensemble building is not affected by ``n_jobs`` but can be controlled by the number
              of models in the ensemble.
            * ``predict()`` is not affected by ``n_jobs`` (in contrast to most scikit-learn models)
            * If ``dask_client`` is ``None``, a new dask client is created.

        dask_client : dask.distributed.Client, optional
            User-created dask client, can be used to start a dask cluster and then
            attach auto-sklearn to it.

        disable_evaluator_output: bool or list, optional (False)
            If True, disable model and prediction output. Cannot be used
            together with ensemble building. ``predict()`` cannot be used when
            setting this True. Can also be used as a list to pass more
            fine-grained information on what to save. Allowed elements in the
            list are:

            * ``'y_optimization'`` : do not save the predictions for the
              optimization set, which would later on be used to build an ensemble.

            * ``model`` : do not save any model files

        smac_scenario_args : dict, optional (None)
            Additional arguments inserted into the scenario of SMAC. See the
            `SMAC documentation <https://automl.github.io/SMAC3/main/api/smac.scenario.scenario.html#module-smac.scenario.scenario>`_
            for a list of available arguments.

        get_smac_object_callback : callable
            Callback function to create an object of class
            `smac.optimizer.smbo.SMBO <https://automl.github.io/SMAC3/main/api/smac.optimizer.smbo.html>`_.
            The function must accept the arguments ``scenario_dict``,
            ``instances``, ``num_params``, ``runhistory``, ``seed`` and ``ta``.
            This is an advanced feature. Use only if you are familiar with
            `SMAC <https://automl.github.io/SMAC3/main/index.html>`_.

        logging_config : dict, optional (None)
            dictionary object specifying the logger configuration. If None,
            the default logging.yaml file is used, which can be found in
            the directory ``util/logging.yaml`` relative to the installation.

        metadata_directory : str, optional (None)
            path to the metadata directory. If None, the default directory
            (autosklearn.metalearning.files) is used.

        metric : Scorer, optional (None)
            An instance of :class:`autosklearn.metrics.Scorer` as created by
            :meth:`autosklearn.metrics.make_scorer`. These are the `Built-in
            Metrics`_.
            If None is provided, a default metric is selected depending on the task.

        scoring_functions : List[Scorer], optional (None)
            List of scorers which will be calculated for each pipeline and results will be
            available via ``cv_results``

        load_models : bool, optional (True)
            Whether to load the models after fitting Auto-sklearn.

        get_trials_callback: callable
            A callable with the following definition.

            * (smac.SMBO, smac.RunInfo, smac.RunValue, time_left: float) -> bool | None

            This will be called after SMAC, the underlying optimizer for autosklearn,
            finishes training each run.

            You can use this to record your own information about the optimization
            process. You can also use this to enable a early stopping based on some
            critera.

            See the example:
            :ref:`Early Stopping And Callbacks <sphx_glr_examples_40_advanced_example_early_stopping_and_callbacks.py>`.

        dataset_compression: Union[bool, Mapping[str, Any]] = True
            We compress datasets so that they fit into some predefined amount of memory.
            Currently this does not apply to dataframes or sparse arrays, only to raw
            numpy arrays.

            **NOTE** - If using a custom ``resampling_strategy`` that relies on specific
            size or ordering of data, this must be disabled to preserve these properties.

            You can disable this entirely by passing ``False`` or leave as the default
            ``True`` for configuration below.

            .. code-block:: python

                {
                    "memory_allocation": 0.1,
                    "methods": ["precision", "subsample"]
                }

            You can also pass your own configuration with the same keys and choosing
            from the available ``"methods"``.

            The available options are described here:

            * **memory_allocation**
                By default, we attempt to fit the dataset into ``0.1 * memory_limit``.
                This float value can be set with ``"memory_allocation": 0.1``.
                We also allow for specifying absolute memory in MB, e.g. 10MB is
                ``"memory_allocation": 10``.

                The memory used by the dataset is checked after each reduction method is
                performed. If the dataset fits into the allocated memory, any further
                methods listed in ``"methods"`` will not be performed.

                For example, if ``methods: ["precision", "subsample"]`` and the
                ``"precision"`` reduction step was enough to make the dataset fit into
                memory, then the ``"subsample"`` reduction step will not be performed.

            * **methods**
                We provide the following methods for reducing the dataset size.
                These can be provided in a list and are performed in the order as given.

                *   ``"precision"`` - We reduce floating point precision as follows:
                    *   ``np.float128 -> np.float64``
                    *   ``np.float96 -> np.float64``
                    *   ``np.float64 -> np.float32``

                *   ``subsample`` - We subsample data such that it **fits directly into
                    the memory allocation** ``memory_allocation * memory_limit``.
                    Therefore, this should likely be the last method listed in
                    ``"methods"``.
                    Subsampling takes into account classification labels and stratifies
                    accordingly. We guarantee that at least one occurrence of each
                    label is included in the sampled set.

        allow_string_features: bool = True
            Whether autosklearn should process string features. By default the
            textpreprocessing is enabled.

        Attributes
        ----------
        cv_results_ : dict of numpy (masked) ndarrays
            A dict with keys as column headers and values as columns, that can be
            imported into a pandas ``DataFrame``.

            Not all keys returned by scikit-learn are supported yet.

        performance_over_time_ : pandas.core.frame.DataFrame
            A ``DataFrame`` containing the models performance over time data. Can be
            used for plotting directly. Please refer to the example
            :ref:`Train and Test Inputs <sphx_glr_examples_40_advanced_example_pandas_train_test.py>`.

        """  # noqa (links are too long)
        # Raise error if the given total time budget is less than 30 seconds.
        if time_left_for_this_task < 30:
            raise ValueError("Time left for this task must be at least " "30 seconds. ")
        self.time_left_for_this_task = time_left_for_this_task
        self.per_run_time_limit = per_run_time_limit
        self.initial_configurations_via_metalearning = (
            initial_configurations_via_metalearning
        )

        # Need to resolve the ensemble class here so we can act on it below.
        if ensemble_class == "default":
            ensemble_class = self._resolve_ensemble_class(metric)

        self.ensemble_class = ensemble_class

        # User specified `ensemble_size` explicitly, warn them about deprecation
        if ensemble_size is not None:
            # Keep consistent behaviour
            message = (
                "`ensemble_size` has been deprecated, please use `ensemble_kwargs = "
                "{'ensemble_size': %d}`. Inserting `ensemble_size` into "
                "`ensemble_kwargs` for now. `ensemble_size` will be removed in "
                "auto-sklearn 0.16."
            ) % ensemble_size
            if ensemble_class == EnsembleSelection:
                if ensemble_kwargs is None:
                    ensemble_kwargs = {"ensemble_size": ensemble_size}
                    warnings.warn(message, DeprecationWarning, stacklevel=2)
                elif "ensemble_size" not in ensemble_kwargs:
                    ensemble_kwargs["ensemble_size"] = ensemble_size
                    warnings.warn(message, DeprecationWarning, stacklevel=2)
                else:
                    warnings.warn(
                        "Deprecated argument `ensemble_size` is both provided "
                        "as an argument to the constructor and passed inside "
                        "`ensemble_kwargs`. Will ignore the argument and use "
                        "the value given in `ensemble_kwargs` (%d). `ensemble_size` "
                        "will be removed in auto-sklearn 0.16."
                        % ensemble_kwargs["ensemble_size"],
                        DeprecationWarning,
                        stacklevel=2,
                    )
            else:
                warnings.warn(
                    "`ensemble_size` has been deprecated, please use "
                    "`ensemble_kwargs = {'ensemble_size': %d} if this "
                    "was intended. Ignoring `ensemble_size` because "
                    "`ensemble_class` != EnsembleSelection. "
                    "`ensemble_size` will be removed in auto-sklearn 0.16."
                    % ensemble_size,
                    DeprecationWarning,
                    stacklevel=2,
                )
        self.ensemble_size = (
            ensemble_size  # Otherwise sklean.base.get_params() will complain
        )
        self.ensemble_kwargs = ensemble_kwargs
        self.ensemble_nbest = ensemble_nbest
        self.max_models_on_disc = max_models_on_disc
        self.seed = seed
        self.memory_limit = memory_limit
        self.include = include
        self.exclude = exclude
        self.resampling_strategy = resampling_strategy
        self.resampling_strategy_arguments = resampling_strategy_arguments
        self.tmp_folder = tmp_folder
        self.delete_tmp_folder_after_terminate = delete_tmp_folder_after_terminate
        self.n_jobs = n_jobs
        self.dask_client = dask_client
        self.disable_evaluator_output = disable_evaluator_output
        self.get_smac_object_callback = get_smac_object_callback
        self.smac_scenario_args = smac_scenario_args
        self.logging_config = logging_config
        self.metadata_directory = metadata_directory
        self.metric = metric
        self.scoring_functions = scoring_functions
        self.load_models = load_models
        self.get_trials_callback = get_trials_callback
        self.dataset_compression = dataset_compression
        self.allow_string_features = allow_string_features

        self.automl_ = None  # type: Optional[AutoML]

        # Handle the number of jobs and the time for them
        self._n_jobs = None
        if self.n_jobs is None or self.n_jobs == 1:
            self._n_jobs = 1
        elif self.n_jobs == -1:
            self._n_jobs = joblib.cpu_count()
        else:
            self._n_jobs = self.n_jobs

        super().__init__()

    def __getstate__(self):
        # Cannot serialize a client!
        self.dask_client = None
        return self.__dict__

    def build_automl(self):

        initial_configs = self.initial_configurations_via_metalearning
        automl = self._get_automl_class()(
            temporary_directory=self.tmp_folder,
            delete_tmp_folder_after_terminate=self.delete_tmp_folder_after_terminate,
            time_left_for_this_task=self.time_left_for_this_task,
            per_run_time_limit=self.per_run_time_limit,
            initial_configurations_via_metalearning=initial_configs,
            ensemble_class=self.ensemble_class,
            ensemble_kwargs=self.ensemble_kwargs,
            ensemble_nbest=self.ensemble_nbest,
            max_models_on_disc=self.max_models_on_disc,
            seed=self.seed,
            memory_limit=self.memory_limit,
            include=self.include,
            exclude=self.exclude,
            resampling_strategy=self.resampling_strategy,
            resampling_strategy_arguments=self.resampling_strategy_arguments,
            n_jobs=self._n_jobs,
            dask_client=self.dask_client,
            get_smac_object_callback=self.get_smac_object_callback,
            disable_evaluator_output=self.disable_evaluator_output,
            smac_scenario_args=self.smac_scenario_args,
            logging_config=self.logging_config,
            metadata_directory=self.metadata_directory,
            metrics=[self.metric] if isinstance(self.metric, Scorer) else self.metric,
            scoring_functions=self.scoring_functions,
            get_trials_callback=self.get_trials_callback,
            dataset_compression=self.dataset_compression,
            allow_string_features=self.allow_string_features,
        )

        return automl

    def fit(self, **kwargs):

        # Automatically set the cutoff time per task
        if self.per_run_time_limit is None:
            self.per_run_time_limit = self._n_jobs * self.time_left_for_this_task // 10

        if self.automl_ is None:
            self.automl_ = self.build_automl()
        self.automl_.fit(load_models=self.load_models, **kwargs)

        return self

    def fit_pipeline(
        self,
        X: SUPPORTED_FEAT_TYPES,
        y: Union[SUPPORTED_TARGET_TYPES, spmatrix],
        config: Union[Configuration, Dict[str, Union[str, float, int]]],
        dataset_name: Optional[str] = None,
        X_test: Optional[SUPPORTED_FEAT_TYPES] = None,
        y_test: Optional[Union[SUPPORTED_TARGET_TYPES, spmatrix]] = None,
        feat_type: Optional[List[str]] = None,
        *args,
        **kwargs: Dict,
    ) -> Tuple[Optional[BasePipeline], RunInfo, RunValue]:
        """Fits and individual pipeline configuration and returns
        the result to the user.

        The Estimator constraints are honored, for example the resampling
        strategy, or memory constraints, unless directly provided to the method.
        By default, this method supports the same signature as fit(), and any extra
        arguments are redirected to the TAE evaluation function, which allows for
        further customization while building a pipeline.

        Any additional argument provided is directly passed to the
        worker exercising the run.

        Parameters
        ----------
        X: array-like, shape = (n_samples, n_features)
            The features used for training
        y: array-like
            The labels used for training
        X_test: Optionalarray-like, shape = (n_samples, n_features)
            If provided, the testing performance will be tracked on this features.
        y_test: array-like
            If provided, the testing performance will be tracked on this labels
        config: Union[Configuration,  Dict[str, Union[str, float, int]]]
            A configuration object used to define the pipeline steps.
            If a dict is passed, a configuration is created based on this dict.
        dataset_name: Optional[str]
            Name that will be used to tag the Auto-Sklearn run and identify the
            Auto-Sklearn run
        feat_type : list, optional (default=None)
            List of str of `len(X.shape[1])` describing the attribute type.
            Possible types are `Categorical` and `Numerical`. `Categorical`
            attributes will be automatically One-Hot encoded. The values
            used for a categorical attribute must be integers, obtained for
            example by `sklearn.preprocessing.LabelEncoder
            <https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.LabelEncoder.html>`_.

        Returns
        -------
        pipeline: Optional[BasePipeline]
            The fitted pipeline. In case of failure while fitting the pipeline,
            a None is returned.
        run_info: RunInFo
            A named tuple that contains the configuration launched
        run_value: RunValue
            A named tuple that contains the result of the run
        """
        if self.automl_ is None:
            self.automl_ = self.build_automl()
        return self.automl_.fit_pipeline(
            X=X,
            y=y,
            dataset_name=dataset_name,
            config=config,
            feat_type=feat_type,
            X_test=X_test,
            y_test=y_test,
            *args,
            **kwargs,
        )

    def fit_ensemble(
        self,
        y,
        task: int = None,
        precision: Literal[16, 21, 64] = 32,
        dataset_name: Optional[str] = None,
        ensemble_size: int | None = None,
        ensemble_kwargs: Optional[Dict[str, Any]] = None,
        ensemble_nbest: Optional[int] = None,
        ensemble_class: Type[AbstractEnsemble] | Literal["default"] | None = "default",
        metric: Scorer | Sequence[Scorer] | None = None,
    ):
        """Fit an ensemble to models trained during an optimization process.

        All parameters are ``None`` by default. If no other value is given,
        the default values which were set in a call to ``fit()`` are used.

        Parameters
        ----------
        y : array-like
            Target values.

        task : int
            A constant from the module ``autosklearn.constants``. Determines
            the task type (binary classification, multiclass classification,
            multilabel classification or regression).

        precision : int
            Numeric precision used when loading ensemble data. Can be either
            ``16``, ``32`` or ``64``.

        dataset_name : str
            Name of the current data set.

        ensemble_size : int, optional
            Number of models added to the ensemble built by *Ensemble
            selection from libraries of models*. Models are drawn with
            replacement. If set to ``0`` no ensemble is fit.

            Deprecated - will be removed in Auto-sklearn 0.16. Please pass
            this argument via ``ensemble_kwargs={"ensemble_size": int}``
            if you want to change the ensemble size for ensemble selection.

        ensemble_kwargs : Dict, optional
            Keyword arguments that are passed to the ensemble class upon
            initialization.

        ensemble_nbest : int
            Only consider the ``ensemble_nbest`` models when building an
            ensemble. This is inspired by a concept called library pruning
            introduced in `Getting Most out of Ensemble Selection`. This
            is independent of the ``ensemble_class`` argument and this
            pruning step is done prior to constructing an ensemble.

        ensemble_class : Type[AbstractEnsemble] | "default", optional (default="default")
            Class implementing the post-hoc ensemble algorithm. Set to
            ``None`` to disable ensemble building or use class:`SingleBest`
            to obtain only use the single best model instead of an
            ensemble.

            If set to "default" it will use :class:`EnsembleSelection` for
            single-objective problems and :class:`MultiObjectiveDummyEnsemble`
            for multi-objective problems.

        metric: Scorer | Sequence[Scorer] | None = None
            A metric or list of metrics to score the ensemble with

        Returns
        -------
        self
        """  # noqa: E501

        if ensemble_class == "default":
            # Things are actually a little more nuanced here:
            # * If they passed `metric=None` at init, we would infer this in automl
            #   during `fit` and store it in the automl instance.
            # * If they passed a `metric` in init and left it `None` here, this would
            #   also be in the automl instance
            # => We can use self.automl_ as ground truth for metric if no metrics passed
            #   and we have one created
            if metric is None and self.automl_ is not None and self.automl_._metrics:
                metric = self.automl_._metrics

            ensemble_class = self._resolve_ensemble_class(metric)

        self.ensemble_class = ensemble_class

        # User specified `ensemble_size` explicitly, warn them about deprecation
        if ensemble_size is not None:
            # Keep consistent behaviour
            message = (
                "`ensemble_size` has been deprecated, please use `ensemble_kwargs = "
                "{'ensemble_size': %d}`. Inserting `ensemble_size` into "
                "`ensemble_kwargs` for now. `ensemble_size` will be removed in "
                "auto-sklearn 0.16."
            ) % ensemble_size
            if ensemble_class == EnsembleSelection:
                if ensemble_kwargs is None:
                    ensemble_kwargs = {"ensemble_size": ensemble_size}
                    warnings.warn(message, DeprecationWarning, stacklevel=2)
                elif "ensemble_size" not in ensemble_kwargs:
                    ensemble_kwargs["ensemble_size"] = ensemble_size
                    warnings.warn(message, DeprecationWarning, stacklevel=2)
                else:
                    warnings.warn(
                        "Deprecated argument `ensemble_size` is both provided "
                        "as an argument to the constructor and passed inside "
                        "`ensemble_kwargs`. Will ignore the argument and use "
                        "the value given in `ensemble_kwargs` (%d). `ensemble_size` "
                        "will be removed in auto-sklearn 0.16."
                        % ensemble_kwargs["ensemble_size"],
                        DeprecationWarning,
                        stacklevel=2,
                    )
            else:
                warnings.warn(
                    "`ensemble_size` has been deprecated, please use "
                    "`ensemble_kwargs = {'ensemble_size': %d} if this "
                    "was intended. Ignoring `ensemble_size` because "
                    "`ensemble_class` != EnsembleSelection. "
                    "`ensemble_size` will be removed in auto-sklearn 0.16."
                    % ensemble_size,
                    DeprecationWarning,
                    stacklevel=2,
                )

        if self.automl_ is None:
            # Build a dummy automl object to call fit_ensemble
            # The ensemble size is honored in the .automl_.fit_ensemble
            # call
            self.automl_ = self.build_automl()
        self.automl_.fit_ensemble(
            y=y,
            task=task,
            precision=precision,
            dataset_name=dataset_name,
            ensemble_nbest=ensemble_nbest,
            ensemble_class=ensemble_class,
            ensemble_kwargs=ensemble_kwargs,
            metrics=metric,
        )
        return self

    def _resolve_ensemble_class(
        self,
        metric: Scorer | Sequence[Scorer] | None,
    ) -> type[AbstractEnsemble]:
        return (
            EnsembleSelection
            if metric is None or isinstance(metric, Scorer) or len(metric) == 1
            else MultiObjectiveDummyEnsemble
        )

    def refit(self, X, y):
        """Refit all models found with fit to new data.

        Necessary when using cross-validation. During training, auto-sklearn
        fits each model k times on the dataset, but does not keep any trained
        model and can therefore not be used to predict for new data points.
        This methods fits all models found during a call to fit on the data
        given. This method may also be used together with holdout to avoid
        only using 66% of the training data to fit the final model.

        Parameters
        ----------

        X : array-like or sparse matrix of shape = [n_samples, n_features]
            The training input samples.

        y : array-like, shape = [n_samples] or [n_samples, n_outputs]
            The targets.

        Returns
        -------

        self

        """
        self.automl_.refit(X, y)
        return self

    def predict(self, X, batch_size=None, n_jobs=1):
        return self.automl_.predict(X, batch_size=batch_size, n_jobs=n_jobs)

    def predict_proba(self, X, batch_size=None, n_jobs=1):
        return self.automl_.predict_proba(X, batch_size=batch_size, n_jobs=n_jobs)

    def score(self, X, y):
        return self.automl_.score(X, y)

    def show_models(self):
        """Returns a dictionary containing dictionaries of ensemble models.

        Each model in the ensemble can be accessed by giving its ``model_id`` as key.

        A model dictionary contains the following:

        * ``"model_id"`` - The id given to a model by ``autosklearn``.

        * ``"rank"`` - The rank of the model based on it's ``"cost"``.

        * ``"cost"`` - The loss of the model on the validation set.

        * ``"ensemble_weight"`` - The weight given to the model in the ensemble.

        * ``"voting_model"`` - The ``cv_voting_ensemble`` model (for 'cv' resampling).

        * ``"estimators"`` - List of models (dicts) in ``cv_voting_ensemble``
            ('cv' resampling).

        * ``"data_preprocessor"`` - The preprocessor used on the data.

        * ``"balancing"`` - The balancing used on the data (for classification).

        * ``"feature_preprocessor"`` - The preprocessor for features types.

        * ``"classifier"`` / ``"regressor"``
          - The autosklearn wrapped classifier or regressor.

        * ``"sklearn_classifier"`` or ``"sklearn_regressor"``
          - The sklearn classifier or regressor.

        **Example**

        .. code-block:: python

            import sklearn.datasets
            import sklearn.metrics
            import autosklearn.regression

            X, y = sklearn.datasets.load_diabetes(return_X_y=True)

            automl = autosklearn.regression.AutoSklearnRegressor(
                time_left_for_this_task=120
                )
            automl.fit(X_train, y_train, dataset_name='diabetes')

            ensemble_dict = automl.show_models()
            print(ensemble_dict)

        Output:

        .. code-block:: text

            {
                25: {'model_id': 25.0,
                     'rank': 1,
                     'cost': 0.43667876507897496,
                     'ensemble_weight': 0.38,
                     'data_preprocessor': <autosklearn.pipeline.components.data_preprocessing....>,
                     'feature_preprocessor': <autosklearn.pipeline.components....>,
                     'regressor': <autosklearn.pipeline.components.regression....>,
                     'sklearn_regressor': SGDRegressor(alpha=0.0006517033225329654,...)
                    },
                6: {'model_id': 6.0,
                    'rank': 2,
                    'cost': 0.4550418898836528,
                    'ensemble_weight': 0.3,
                    'data_preprocessor': <autosklearn.pipeline.components.data_preprocessing....>,
                    'feature_preprocessor': <autosklearn.pipeline.components....>,
                    'regressor': <autosklearn.pipeline.components.regression....>,
                    'sklearn_regressor': ARDRegression(alpha_1=0.0003701926442639788,...)
                    }...
            }

        Returns
        -------
        Dict(int, Any) : dictionary of length = number of models in the ensemble
            A dictionary of models in the ensemble, where ``model_id`` is the key.

        """  # noqa: E501

        return self.automl_.show_models()

    def get_models_with_weights(self):
        """Return a list of the final ensemble found by auto-sklearn.

        Returns
        -------
        [(weight_1, model_1), ..., (weight_n, model_n)]

        """
        return self.automl_.get_models_with_weights()

    @property
    def performance_over_time_(self):
        return self.automl_.performance_over_time_

    @property
    def cv_results_(self):
        return self.automl_.cv_results_

    @property
    def trajectory_(self):
        return self.automl_.trajectory_

    @property
    def fANOVA_input_(self):
        return self.automl_.fANOVA_input_

    def sprint_statistics(self):
        """Return the following statistics of the training result:

        - dataset name
        - metric used
        - best validation score
        - number of target algorithm runs
        - number of successful target algorithm runs
        - number of crashed target algorithm runs
        - number of target algorithm runs that exceeded the memory limit
        - number of target algorithm runs that exceeded the time limit

        Returns
        -------
        str
        """
        return self.automl_.sprint_statistics()

    def leaderboard(
        self,
        detailed: bool = False,
        ensemble_only: bool = True,
        top_k: Union[int, Literal["all"]] = "all",
        sort_by: str = "cost",
        sort_order: Literal["auto", "ascending", "descending"] = "auto",
        include: Optional[Union[str, Iterable[str]]] = None,
    ) -> pd.DataFrame:
        """Returns a pandas table of results for all evaluated models.

        Gives an overview of all models trained during the search process along
        with various statistics about their training.

        The available statistics are:

        **Simple**:

        * ``"model_id"`` - The id given to a model by ``autosklearn``.
        * ``"rank"`` - The rank of the model based on it's ``"cost"``.
        * ``"ensemble_weight"`` - The weight given to the model in the ensemble.
        * ``"type"`` - The type of classifier/regressor used.
        * ``"cost"`` - The loss of the model on the validation set.
        * ``"duration"`` - Length of time the model was optimized for.

        **Detailed**:
        The detailed view includes all of the simple statistics along with the
        following.

        * ``"config_id"`` - The id used by SMAC for optimization.
        * ``"budget"`` - How much budget was allocated to this model.
        * ``"status"`` - The return status of training the model with SMAC.
        * ``"train_loss"`` - The loss of the model on the training set.
        * ``"balancing_strategy"`` - The balancing strategy used for data preprocessing.
        * ``"start_time"`` - Time the model began being optimized
        * ``"end_time"`` - Time the model ended being optimized
        * ``"data_preprocessors"`` - The preprocessors used on the data
        * ``"feature_preprocessors"`` - The preprocessors for features types

        Parameters
        ----------
        detailed: bool = False
            Whether to give detailed information or just a simple overview.

        ensemble_only: bool = True
            Whether to view only models included in the ensemble or all models
            trained.

        top_k: int or "all" = "all"
            How many models to display.

        sort_by: str = 'cost'
            What column to sort by. If that column is not present, the
            sorting defaults to the ``"model_id"`` index column.

            Defaults to the metric optimized. Sort by the first objective
            in case of a multi-objective optimization problem

        sort_order: "auto" or "ascending" or "descending" = "auto"
            Which sort order to apply to the ``sort_by`` column. If left
            as ``"auto"``, it will sort by a sensible default where "better" is
            on top, otherwise defaulting to the pandas default for
            `DataFrame.sort_values`_ if there is no obvious "better".

            .. _DataFrame.sort_values: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.sort_values.html

        include: Optional[str or Iterable[str]]
            Items to include, other items not specified will be excluded.
            The exception is the ``"model_id"`` index column which is always included.

            If left as ``None``, it will resort back to using the ``detailed``
            param to decide the columns to include.

        Returns
        -------
        pd.DataFrame
            A dataframe of statistics for the models, ordered by ``sort_by``.

        """  # noqa (links are too long)
        # TODO validate that `self` is fitted. This is required for
        #      self.ensemble_ to get the identifiers of models it will generate
        #      weights for.
        num_metrics = (
            1
            if self.metric is None or isinstance(self.metric, Scorer)
            else len(self.metric)
        )
        column_types = AutoSklearnEstimator._leaderboard_columns(num_metrics)
        if num_metrics == 1:
            multi_objective_cost_names = []
        else:
            multi_objective_cost_names = [f"cost_{i}" for i in range(num_metrics)]

        # Validation of top_k
        if (
            not (isinstance(top_k, str) or isinstance(top_k, int))
            or (isinstance(top_k, str) and top_k != "all")
            or (isinstance(top_k, int) and top_k <= 0)
        ):
            raise ValueError(
                f"top_k={top_k} must be a positive integer or pass"
                " `top_k`='all' to view results for all models"
            )

        # Validate columns to include
        if isinstance(include, str):
            include = [include]

        if include == ["model_id"]:
            raise ValueError("Must provide more than just `model_id`")

        if include is not None:
            columns = [*include]

            # 'model_id' should always be present as it is the unique index
            # used for pandas
            if "model_id" not in columns:
                columns.append("model_id")

            invalid_include_items = set(columns) - set(column_types["all"])
            if len(invalid_include_items) != 0:
                raise ValueError(
                    f"Values {invalid_include_items} are not known"
                    f" columns to include, must be contained in "
                    f"{column_types['all']}"
                )
        elif detailed:
            columns = column_types["all"]
        else:
            columns = column_types["simple"]

        # Validation of sorting
        if sort_by == "cost":
            sort_by_cost = True
            if num_metrics == 1:
                sort_by = ["cost", "model_id"]
            else:
                sort_by = multi_objective_cost_names + ["model_id"]
        else:
            sort_by_cost = False
            if isinstance(sort_by, str):
                if sort_by not in column_types["all"]:
                    raise ValueError(
                        f"sort_by='{sort_by}' must be one of included "
                        f"columns {set(column_types['all'])}"
                    )
            elif len(set(sort_by) - set(column_types["all"])) > 0:
                too_much = set(sort_by) - set(column_types["all"])
                raise ValueError(
                    f"sort_by='{too_much}' must be in the included columns "
                    f"{set(column_types['all'])}"
                )

        valid_sort_orders = ["auto", "ascending", "descending"]
        if not (isinstance(sort_order, str) and sort_order in valid_sort_orders):
            raise ValueError(
                f"`sort_order` = {sort_order} must be a str in " f"{valid_sort_orders}"
            )

        # To get all the models that were optmized, we collect what we can from
        # runhistory first.
        def additional_info_has_key(rv, key):
            return rv.additional_info and key in rv.additional_info

        model_runs = {}
        for run_key, run_val in self.automl_.runhistory_.data.items():
            if not additional_info_has_key(run_val, "num_run"):
                continue
            else:
                model_key = run_val.additional_info["num_run"]
                model_run = {
                    "model_id": run_val.additional_info["num_run"],
                    "seed": run_key.seed,
                    "budget": run_key.budget,
                    "duration": run_val.time,
                    "config_id": run_key.config_id,
                    "start_time": run_val.starttime,
                    "end_time": run_val.endtime,
                    "status": str(run_val.status),
                    "train_loss": run_val.additional_info["train_loss"]
                    if additional_info_has_key(run_val, "train_loss")
                    else None,
                    "config_origin": run_val.additional_info["configuration_origin"]
                    if additional_info_has_key(run_val, "configuration_origin")
                    else None,
                }
                if num_metrics == 1:
                    model_run["cost"] = run_val.cost
                else:
                    for cost_idx, cost in enumerate(run_val.cost):
                        model_run[f"cost_{cost_idx}"] = cost
                model_runs[model_key] = model_run

        # Next we get some info about the model itself
        model_class_strings = {
            AutoMLClassifier: "classifier",
            AutoMLRegressor: "regressor",
        }
        model_type = model_class_strings.get(self._get_automl_class(), None)
        if model_type is None:
            raise RuntimeError(f"Unknown `automl_class` {self._get_automl_class()}")

        # A dict mapping model ids to their configurations
        configurations = self.automl_.runhistory_.ids_config

        for model_id, run_info in model_runs.items():
            config_id = run_info["config_id"]
            run_config = configurations[config_id]._values

            run_info.update(
                {
                    "balancing_strategy": run_config.get("balancing:strategy", None),
                    "type": run_config[f"{model_type}:__choice__"],
                    "data_preprocessors": [
                        value
                        for key, value in run_config.items()
                        if "data_preprocessing" in key and "__choice__" in key
                    ],
                    "feature_preprocessors": [
                        value
                        for key, value in run_config.items()
                        if "feature_preprocessor" in key and "__choice__" in key
                    ],
                }
            )

        # Get the models ensemble weight if it has one
        for (
            _,
            model_id,
            _,
        ), weight in self.automl_.ensemble_.get_identifiers_with_weights():

            # We had issues where the model's in the ensembles are not in the runhistory
            # collected. I have no clue why this is but to prevent failures, we fill
            # the values with NaN
            if model_id not in model_runs:
                model_run = {
                    "model_id": model_id,
                    "seed": pd.NA,
                    "budget": pd.NA,
                    "duration": pd.NA,
                    "config_id": pd.NA,
                    "start_time": pd.NA,
                    "end_time": pd.NA,
                    "status": pd.NA,
                    "train_loss": pd.NA,
                    "config_origin": pd.NA,
                    "type": pd.NA,
                }
                if num_metrics == 1:
                    model_run["cost"] = pd.NA
                else:
                    for cost_idx in range(num_metrics):
                        model_run[f"cost_{cost_idx}"] = pd.NA
                model_runs[model_id] = model_run

            model_runs[model_id]["ensemble_weight"] = weight

        # Filter out non-ensemble members if needed, else fill in a default
        # value of 0 if it's missing
        if ensemble_only:
            model_runs = {
                model_id: info
                for model_id, info in model_runs.items()
                if ("ensemble_weight" in info and info["ensemble_weight"] > 0)
            }
        else:
            for model_id, info in model_runs.items():
                if "ensemble_weight" not in info:
                    info["ensemble_weight"] = 0

        # `rank` relies on `cost` so we include `cost`
        # We drop it later if it's not requested
        if "rank" in columns:
            if num_metrics == 1 and "cost" not in columns:
                columns = [*columns, "cost"]
            elif num_metrics > 1 and any(
                cost_name not in columns for cost_name in multi_objective_cost_names
            ):
                columns = columns + list(multi_objective_cost_names)

        # Finally, convert into a tabular format by converting the dict into
        # column wise orientation.
        dataframe = pd.DataFrame(
            {
                col: [run_info[col] for run_info in model_runs.values()]
                for col in columns
                if col != "rank"
            }
        )

        # Give it an index, even if not in the `include`
        dataframe.set_index("model_id", inplace=True)

        # Add the `rank` column if needed
        # requested by the user
        if "rank" in columns:
            if num_metrics == 1:
                dataframe.sort_values(by="cost", ascending=True, inplace=True)
            else:
                dataframe.sort_values(by="cost_0", ascending=True, inplace=True)
            dataframe.insert(
                column="rank",
                value=range(1, len(dataframe) + 1),
                loc=list(columns).index("rank") - 1,
            )  # account for `model_id`

        # Decide on the sort order depending on what it gets sorted by
        descending_columns = ["ensemble_weight", "duration"]
        if sort_order == "auto":
            ascending_param = [
                False if sby in descending_columns else True for sby in sort_by
            ]
        else:
            ascending_param = False if sort_order == "descending" else True

        # Sort by the given column name, defaulting to 'model_id' if not present
        if (
            (not sort_by_cost and len(set(sort_by) - set(dataframe.columns)) > 0)
            or (sort_by_cost and "cost" not in dataframe.columns)
            or (
                sort_by_cost
                and any(
                    cost_name not in dataframe.columns
                    for cost_name in multi_objective_cost_names
                )
            )
        ):
            self.automl_._logger.warning(
                f"sort_by = '{sort_by}' was not present"
                ", defaulting to sort on the index "
                "'model_id'"
            )
            sort_by = "model_id"
            sort_by_cost = False
            ascending_param = True

        # Single objective
        if sort_by_cost:
            dataframe.sort_values(
                by=sort_by, ascending=[True] * len(sort_by), inplace=True
            )
        else:
            dataframe.sort_values(by=sort_by, ascending=ascending_param, inplace=True)

        if num_metrics == 1:
            if "cost" not in columns and "cost" in dataframe.columns:
                dataframe.drop("cost", inplace=True)
        else:
            for cost_name in multi_objective_cost_names:
                if cost_name not in columns and cost_name in dataframe.columns:
                    dataframe.drop(cost_name, inplace=True)

        # Lastly, just grab the top_k
        if top_k == "all" or top_k >= len(dataframe):
            top_k = len(dataframe)

        dataframe = dataframe.head(top_k)

        return dataframe

    @staticmethod
    def _leaderboard_columns(
        num_metrics: int,
    ) -> Dict[Literal["all", "simple", "detailed"], List[str]]:
        if num_metrics == 1:
            cost_list = ["cost"]
        else:
            cost_list = [f"cost_{i}" for i in range(num_metrics)]
        all = (
            [
                "model_id",
                "rank",
                "ensemble_weight",
                "type",
            ]
            + cost_list
            + [
                "duration",
                "config_id",
                "train_loss",
                "seed",
                "start_time",
                "end_time",
                "budget",
                "status",
                "data_preprocessors",
                "feature_preprocessors",
                "balancing_strategy",
                "config_origin",
            ]
        )
        simple = (
            ["model_id", "rank", "ensemble_weight", "type"] + cost_list + ["duration"]
        )
        detailed = all
        return {"all": all, "detailed": detailed, "simple": simple}

    def _get_automl_class(self):
        raise NotImplementedError()

    def get_configuration_space(
        self,
        X: SUPPORTED_FEAT_TYPES,
        y: Union[SUPPORTED_TARGET_TYPES, spmatrix],
        X_test: Optional[SUPPORTED_FEAT_TYPES] = None,
        y_test: Optional[Union[SUPPORTED_TARGET_TYPES, spmatrix]] = None,
        dataset_name: Optional[str] = None,
        feat_type: Optional[List[str]] = None,
    ) -> ConfigurationSpace:
        """
        Returns the Configuration Space object, from which Auto-Sklearn
        will sample configurations and build pipelines.

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]
            Array with the training features, used to get characteristics like
            data sparsity
        y : array-like, shape = [n_samples] or [n_samples, n_outputs]
            Array with the problem labels
        X_test : array-like or sparse matrix of shape = [n_samples, n_features]
            Array with features used for performance estimation
        y_test : array-like, shape = [n_samples] or [n_samples, n_outputs]
            Array with the problem labels for the testing split
        dataset_name: Optional[str]
            A string to tag the Auto-Sklearn run
        """
        if self.automl_ is None:
            self.automl_ = self.build_automl()

        return (
            self.automl_.fit(
                X,
                y,
                X_test=X_test,
                y_test=y_test,
                dataset_name=dataset_name,
                feat_type=feat_type,
                only_return_configuration_space=True,
            )
            if self.automl_.configuration_space is None
            else self.automl_.configuration_space
        )

    def get_pareto_set(self) -> Sequence[VotingClassifier | VotingRegressor]:
        return self.automl_._load_pareto_set()


class AutoSklearnClassifier(AutoSklearnEstimator, ClassifierMixin):
    """This class implements the classification task."""

    def fit(self, X, y, X_test=None, y_test=None, feat_type=None, dataset_name=None):
        """Fit *auto-sklearn* to given training set (X, y).

        Fit both optimizes the machine learning models and builds an ensemble
        out of them.

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]
            The training input samples.

        y : array-like, shape = [n_samples] or [n_samples, n_outputs]
            The target classes.

        X_test : array-like or sparse matrix of shape = [n_samples, n_features]
            Test data input samples. Will be used to save test predictions for
            all models. This allows to evaluate the performance of Auto-sklearn
            over time.

        y_test : array-like, shape = [n_samples] or [n_samples, n_outputs]
            Test data target classes. Will be used to calculate the test error
            of all models. This allows to evaluate the performance of
            Auto-sklearn over time.

        feat_type : list, optional (default=None)
            List of str of `len(X.shape[1])` describing the attribute type.
            Possible types are `Categorical` and `Numerical`. `Categorical`
            attributes will be automatically One-Hot encoded. The values
            used for a categorical attribute must be integers, obtained for
            example by `sklearn.preprocessing.LabelEncoder
            <https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.LabelEncoder.html>`_.

        dataset_name : str, optional (default=None)
            Create nicer output. If None, a string will be determined by the
            md5 hash of the dataset.

        Returns
        -------
        self
        """
        # AutoSklearn does not handle sparse y for now
        y = convert_if_sparse(y)

        # Before running anything else, first check that the
        # type of data is compatible with auto-sklearn. Legal target
        # types are: binary, multiclass, multilabel-indicator.
        target_type = type_of_target(y)
        supported_types = ["binary", "multiclass", "multilabel-indicator"]
        if target_type not in supported_types:
            raise ValueError(
                "Classification with data of type {} is "
                "not supported. Supported types are {}. "
                "You can find more information about scikit-learn "
                "data types in: "
                "https://scikit-learn.org/stable/modules/multiclass.html"
                "".format(target_type, supported_types)
            )

        # remember target type for using in predict_proba later.
        self.target_type = target_type

        super().fit(
            X=X,
            y=y,
            X_test=X_test,
            y_test=y_test,
            feat_type=feat_type,
            dataset_name=dataset_name,
        )

        # After fit, a classifier is expected to define classes_
        # A list of class labels known to the classifier, mapping each label
        # to a numerical index used in the model representation our output.
        self.classes_ = self.automl_.InputValidator.target_validator.classes_

        return self

    def predict(self, X, batch_size=None, n_jobs=1):
        """Predict classes for X.

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]

        Returns
        -------
        y : array of shape = [n_samples] or [n_samples, n_labels]
            The predicted classes.
        """
        return super().predict(X, batch_size=batch_size, n_jobs=n_jobs)

    def predict_proba(self, X, batch_size=None, n_jobs=1):
        """Predict probabilities of classes for all samples X.

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]

        batch_size : int (optional)
            Number of data points to predict for (predicts all points at once
            if ``None``.
        n_jobs : int

        Returns
        -------
        y : array of shape = [n_samples, n_classes] or [n_samples, n_labels]
            The predicted class probabilities.
        """
        pred_proba = super().predict_proba(X, batch_size=batch_size, n_jobs=n_jobs)

        # Check if all probabilities sum up to 1.
        # Assert only if target type is not multilabel-indicator.
        if self.target_type not in ["multilabel-indicator"]:
            assert np.allclose(
                np.sum(pred_proba, axis=1), np.ones_like(pred_proba[:, 0])
            ), "prediction probability does not sum up to 1!"

        # Check that all probability values lie between 0 and 1.
        assert (pred_proba >= 0).all() and (
            pred_proba <= 1
        ).all(), "found prediction probability value outside of [0, 1]!"

        return pred_proba

    def _get_automl_class(self):
        return AutoMLClassifier


class AutoSklearnRegressor(AutoSklearnEstimator, RegressorMixin):
    """
    This class implements the regression task.

    """

    def fit(self, X, y, X_test=None, y_test=None, feat_type=None, dataset_name=None):
        """Fit *Auto-sklearn* to given training set (X, y).

        Fit both optimizes the machine learning models and builds an ensemble
        out of them.

        Parameters
        ----------

        X : array-like or sparse matrix of shape = [n_samples, n_features]
            The training input samples.

        y : array-like, shape = [n_samples] or [n_samples, n_targets]
            The regression target.

        X_test : array-like or sparse matrix of shape = [n_samples, n_features]
            Test data input samples. Will be used to save test predictions for
            all models. This allows to evaluate the performance of Auto-sklearn
            over time.

        y_test : array-like, shape = [n_samples] or [n_samples, n_targets]
            The regression target. Will be used to calculate the test error
            of all models. This allows to evaluate the performance of
            Auto-sklearn over time.

        feat_type : list, optional (default=None)
            List of str of `len(X.shape[1])` describing the attribute type.
            Possible types are `Categorical` and `Numerical`. `Categorical`
            attributes will be automatically One-Hot encoded.

        dataset_name : str, optional (default=None)
            Create nicer output. If None, a string will be determined by the
            md5 hash of the dataset.

        Returns
        -------
        self

        """
        # Before running anything else, first check that the
        # type of data is compatible with auto-sklearn. Legal target
        # types are: continuous, continuous-multioutput, and the special cases:
        # multiclass : because [3.0, 1.0, 5.0] is considered as multiclass
        # binary: because [1.0, 0.0] is considered multiclass
        # AutoSklearn does not handle sparse y for now
        y = convert_if_sparse(y)

        target_type = type_of_target(y)
        supported_types = [
            "continuous",
            "binary",
            "multiclass",
            "continuous-multioutput",
        ]
        if target_type not in supported_types:
            raise ValueError(
                "Regression with data of type {} is "
                "not supported. Supported types are {}. "
                "You can find more information about scikit-learn "
                "data types in: "
                "https://scikit-learn.org/stable/modules/multiclass.html"
                "".format(target_type, supported_types)
            )

        # Fit is supposed to be idempotent!
        # But not if we use share_mode.
        super().fit(
            X=X,
            y=y,
            X_test=X_test,
            y_test=y_test,
            feat_type=feat_type,
            dataset_name=dataset_name,
        )

        return self

    def predict(self, X, batch_size=None, n_jobs=1):
        """Predict regression target for X.

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]

        Returns
        -------
        y : array of shape = [n_samples] or [n_samples, n_outputs]
            The predicted values.

        """
        return super().predict(X, batch_size=batch_size, n_jobs=n_jobs)

    def _get_automl_class(self):
        return AutoMLRegressor



from typing import Optional

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnClassificationAlgorithm
from autosklearn.pipeline.constants import DENSE, PREDICTIONS, SIGNED_DATA, SPARSE
from autosklearn.util.common import check_for_bool


class MultinomialNB(AutoSklearnClassificationAlgorithm):
    def __init__(self, alpha, fit_prior, random_state=None, verbose=0):
        self.alpha = alpha
        self.fit_prior = fit_prior
        self.random_state = random_state
        self.verbose = int(verbose)
        self.estimator = None

    def fit(self, X, y):
        import scipy.sparse
        import sklearn.naive_bayes

        self.fit_prior = check_for_bool(self.fit_prior)
        self.alpha = float(self.alpha)
        self.n_iter = 0
        self.fully_fit_ = False
        self.estimator = sklearn.naive_bayes.MultinomialNB(
            alpha=self.alpha,
            fit_prior=self.fit_prior,
        )
        self.classes_ = np.unique(y.astype(int))

        # Because the pipeline guarantees that each feature is positive,
        # clip all values below zero to zero
        if scipy.sparse.issparse(X):
            X.data[X.data < 0] = 0.0
        else:
            X[X < 0] = 0.0

        # Fallback for multilabel classification
        if len(y.shape) > 1 and y.shape[1] > 1:
            import sklearn.multiclass

            self.estimator = sklearn.multiclass.OneVsRestClassifier(
                self.estimator, n_jobs=1
            )
        self.estimator.fit(X, y)

        return self

    def predict(self, X):
        if self.estimator is None:
            raise NotImplementedError
        return self.estimator.predict(X)

    def predict_proba(self, X):
        if self.estimator is None:
            raise NotImplementedError()
        return self.estimator.predict_proba(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "MultinomialNB",
            "name": "Multinomial Naive Bayes classifier",
            "handles_regression": False,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": False,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, SIGNED_DATA),
            "output": (PREDICTIONS,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        cs = ConfigurationSpace()

        # the smoothing parameter is a non-negative float
        # I will limit it to 100 and put it on a logarithmic scale. (SF)
        # Please adjust that, if you know a proper range, this is just a guess.
        alpha = UniformFloatHyperparameter(
            name="alpha", lower=1e-2, upper=100, default_value=1, log=True
        )

        fit_prior = CategoricalHyperparameter(
            name="fit_prior", choices=["True", "False"], default_value="True"
        )

        cs.add_hyperparameters([alpha, fit_prior])

        return cs



from collections import OrderedDict

import numpy as np
import pandas as pd
from ConfigSpace.configuration_space import Configuration
from ConfigSpace.util import deactivate_inactive_hyperparameters

from ..input import aslib_simple
from ..metafeatures.metafeature import DatasetMetafeatures


class Run(object):
    def __init__(self, configuration, result, runtime):
        self.configuration = configuration
        self.result = result
        self.runtime = runtime

    def __repr__(self):
        return "Run:\nresult: %3.3f\nruntime: %3.3f\n%s" % (
            self.result,
            self.runtime,
            str(self.configuration),
        )


class Instance(object):
    def __init__(self, name, features):
        self.name = name
        self.features = features


class MetaBase(object):
    def __init__(self, configuration_space, aslib_directory, logger):
        """Container for dataset metadata and experiment results.

        Constructor arguments:
        - The configuration space
        - aslib_directory: directory with a problem instance in the aslib format
        """

        self.logger = logger

        self.configuration_space = configuration_space
        self.default_configuration_space_dict = (
            configuration_space.get_default_configuration().get_dictionary()
        )
        self.aslib_directory = aslib_directory

        aslib_reader = aslib_simple.AlgorithmSelectionProblem(
            self.aslib_directory, self.configuration_space
        )
        self.metafeatures = aslib_reader.metafeatures
        self.algorithm_runs: OrderedDict[
            str, pd.DataFrame
        ] = aslib_reader.algorithm_runs
        self.configurations = aslib_reader.configurations

        configurations = dict()
        for algorithm_id in self.configurations:
            configuration = self.configurations[algorithm_id]
            try:
                for key in self.default_configuration_space_dict.keys():
                    if key not in configuration:
                        configuration[key] = self.default_configuration_space_dict[key]
                configuration = Configuration(
                    configuration_space,
                    values=configuration,
                    allow_inactive_with_values=True,
                )
                configuration = deactivate_inactive_hyperparameters(
                    configuration, configuration_space
                )
                configurations[str(algorithm_id)] = configuration
            except (ValueError, KeyError) as e:
                self.logger.debug("Error reading configurations: %s", e)

        self.configurations = configurations

    def add_dataset(self, name, metafeatures):
        metafeatures.name = name
        if isinstance(metafeatures, DatasetMetafeatures):
            data_ = {
                mf.name: mf.value for mf in metafeatures.metafeature_values.values()
            }
            metafeatures = pd.Series(name=name, data=data_, dtype=np.float64)
        if name.lower() in self.metafeatures.index:
            self.logger.warning(
                "Dataset %s already in meta-data. Removing occurence.", name.lower()
            )
            self.metafeatures.drop(name.lower(), inplace=True)
        self.metafeatures = pd.concat([self.metafeatures, pd.DataFrame(metafeatures).T])

    def get_runs(self, dataset_name, performance_measure=None):
        """Return a list of all runs for a dataset."""
        if performance_measure is None:
            performance_measure = list(self.algorithm_runs.keys())[0]
        return self.algorithm_runs[performance_measure].loc[dataset_name]

    def get_all_runs(self, performance_measure=None):
        """Return a dictionary with a list of all runs"""
        if performance_measure is None:
            performance_measure = list(self.algorithm_runs.keys())[0]
        return self.algorithm_runs[performance_measure]

    def get_metafeatures(self, dataset_name=None, features=None):
        if features is not None:
            metafeatures = self._get_metafeatures(features)
        else:
            metafeatures = self.metafeatures
        if dataset_name is not None:
            return metafeatures.loc[dataset_name]
        else:
            return metafeatures

    def _get_metafeatures(self, features):
        """This is inside an extra function for testing purpose"""
        # Load the task

        self.logger.info("Going to use the following metafeature subset: %s", features)
        all_metafeatures = self.metafeatures
        all_metafeatures = all_metafeatures.loc[:, features]

        return all_metafeatures

    def get_configuration_from_algorithm_index(self, idx):
        return self.configurations[str(idx)]
        # configuration = self.configurations[idx]
        # configuration = Configuration(self.configuration_space,
        # **configuration)
        # return configuration

    def get_algorithm_index_from_configuration(self, configuration):
        for idx in self.configurations.keys():
            if configuration == self.configurations[idx]:
                return idx

        raise ValueError(configuration)

    def get_all_dataset_names(self):
        return list(self.metafeatures.index)



import sklearn.svm

from autosklearn.pipeline.components.classification.liblinear_svc import LibLinear_SVC

from .test_base import BaseClassificationComponentTest


class LibLinearComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 1
    res["default_iris_iterative"] = -1
    res["default_iris_proba"] = 0.3350793047400861
    res["default_iris_sparse"] = 0.56
    res["default_digits"] = 0.914996964177292
    res["default_digits_places"] = 2
    res["default_digits_iterative"] = -1
    res["default_digits_binary"] = 0.98907103825136611
    res["default_digits_multilabel"] = 0.89889188078944637
    res["default_digits_multilabel_places"] = 2
    res["default_digits_multilabel_proba"] = 0.99999999999999989

    sk_mod = sklearn.svm.LinearSVC
    module = LibLinear_SVC



from typing import Optional

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnRegressionAlgorithm
from autosklearn.pipeline.constants import DENSE, PREDICTIONS, SPARSE, UNSIGNED_DATA


class AdaboostRegressor(AutoSklearnRegressionAlgorithm):
    def __init__(self, n_estimators, learning_rate, loss, max_depth, random_state=None):
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.loss = loss
        self.random_state = random_state
        self.max_depth = max_depth
        self.estimator = None

    def fit(self, X, y):
        import sklearn.ensemble
        import sklearn.tree

        self.n_estimators = int(self.n_estimators)
        self.learning_rate = float(self.learning_rate)
        self.max_depth = int(self.max_depth)
        base_estimator = sklearn.tree.DecisionTreeRegressor(max_depth=self.max_depth)

        self.estimator = sklearn.ensemble.AdaBoostRegressor(
            base_estimator=base_estimator,
            n_estimators=self.n_estimators,
            learning_rate=self.learning_rate,
            loss=self.loss,
            random_state=self.random_state,
        )

        if y.ndim == 2 and y.shape[1] == 1:
            y = y.flatten()

        self.estimator.fit(X, y)
        return self

    def predict(self, X):
        if self.estimator is None:
            raise NotImplementedError
        return self.estimator.predict(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "AB",
            "name": "AdaBoost Regressor",
            "handles_regression": True,
            "handles_classification": False,
            "handles_multiclass": False,
            "handles_multilabel": False,
            "handles_multioutput": False,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (PREDICTIONS,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        cs = ConfigurationSpace()

        # base_estimator = Constant(name="base_estimator", value="None")
        n_estimators = UniformIntegerHyperparameter(
            name="n_estimators", lower=50, upper=500, default_value=50, log=False
        )
        learning_rate = UniformFloatHyperparameter(
            name="learning_rate", lower=0.01, upper=2, default_value=0.1, log=True
        )
        loss = CategoricalHyperparameter(
            name="loss",
            choices=["linear", "square", "exponential"],
            default_value="linear",
        )
        max_depth = UniformIntegerHyperparameter(
            name="max_depth", lower=1, upper=10, default_value=1, log=False
        )

        cs.add_hyperparameters([n_estimators, learning_rate, loss, max_depth])
        return cs



from typing import Optional

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.forbidden import ForbiddenAndConjunction, ForbiddenEqualsClause
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    Constant,
    UniformFloatHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnRegressionAlgorithm
from autosklearn.pipeline.constants import DENSE, PREDICTIONS, SPARSE, UNSIGNED_DATA
from autosklearn.util.common import check_for_bool


class LibLinear_SVR(AutoSklearnRegressionAlgorithm):
    # Liblinear is not deterministic as it uses a RNG inside
    def __init__(
        self,
        loss,
        epsilon,
        dual,
        tol,
        C,
        fit_intercept,
        intercept_scaling,
        random_state=None,
    ):
        self.epsilon = epsilon
        self.loss = loss
        self.dual = dual
        self.tol = tol
        self.C = C
        self.fit_intercept = fit_intercept
        self.intercept_scaling = intercept_scaling
        self.random_state = random_state
        self.estimator = None

    def fit(self, X, y):
        import sklearn.svm

        self.C = float(self.C)
        self.tol = float(self.tol)
        self.epsilon = float(self.epsilon)

        self.dual = check_for_bool(self.dual)
        self.fit_intercept = check_for_bool(self.fit_intercept)
        self.intercept_scaling = float(self.intercept_scaling)

        self.estimator = sklearn.svm.LinearSVR(
            epsilon=self.epsilon,
            loss=self.loss,
            dual=self.dual,
            tol=self.tol,
            C=self.C,
            fit_intercept=self.fit_intercept,
            intercept_scaling=self.intercept_scaling,
            random_state=self.random_state,
        )

        if y.ndim == 2 and y.shape[1] == 1:
            y = y.flatten()

        self.estimator.fit(X, y)
        return self

    def predict(self, X):
        if self.estimator is None:
            raise NotImplementedError()
        return self.estimator.predict(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "Liblinear-SVR",
            "name": "Liblinear Support Vector Regression",
            "handles_regression": True,
            "handles_classification": False,
            "handles_multiclass": False,
            "handles_multilabel": False,
            "handles_multioutput": False,
            "is_deterministic": False,
            "input": (SPARSE, DENSE, UNSIGNED_DATA),
            "output": (PREDICTIONS,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        cs = ConfigurationSpace()
        C = UniformFloatHyperparameter("C", 0.03125, 32768, log=True, default_value=1.0)
        loss = CategoricalHyperparameter(
            "loss",
            ["epsilon_insensitive", "squared_epsilon_insensitive"],
            default_value="squared_epsilon_insensitive",
        )
        # Random Guess
        epsilon = UniformFloatHyperparameter(
            name="epsilon", lower=0.001, upper=1, default_value=0.1, log=True
        )
        dual = Constant("dual", "False")
        # These are set ad-hoc
        tol = UniformFloatHyperparameter(
            "tol", 1e-5, 1e-1, default_value=1e-4, log=True
        )
        fit_intercept = Constant("fit_intercept", "True")
        intercept_scaling = Constant("intercept_scaling", 1)

        cs.add_hyperparameters(
            [C, loss, epsilon, dual, tol, fit_intercept, intercept_scaling]
        )

        dual_and_loss = ForbiddenAndConjunction(
            ForbiddenEqualsClause(dual, "False"),
            ForbiddenEqualsClause(loss, "epsilon_insensitive"),
        )
        cs.add_forbidden_clause(dual_and_loss)

        return cs



from typing import Optional

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import UniformFloatHyperparameter

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnClassificationAlgorithm
from autosklearn.pipeline.constants import DENSE, PREDICTIONS, UNSIGNED_DATA
from autosklearn.pipeline.implementations.util import softmax


class QDA(AutoSklearnClassificationAlgorithm):
    def __init__(self, reg_param, random_state=None):
        self.reg_param = float(reg_param)
        self.estimator = None

    def fit(self, X, Y):
        import sklearn.discriminant_analysis

        estimator = sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis(
            reg_param=self.reg_param
        )

        if len(Y.shape) == 2 and Y.shape[1] > 1:
            import sklearn.multiclass

            self.estimator = sklearn.multiclass.OneVsRestClassifier(estimator, n_jobs=1)
        else:
            self.estimator = estimator

        self.estimator.fit(X, Y)

        if len(Y.shape) == 2 and Y.shape[1] > 1:
            problems = []
            for est in self.estimator.estimators_:
                problem = np.any(np.any([np.any(s <= 0.0) for s in est.scalings_]))
                problems.append(problem)
            problem = np.any(problems)
        else:
            problem = np.any(
                np.any([np.any(s <= 0.0) for s in self.estimator.scalings_])
            )
        if problem:
            raise ValueError(
                "Numerical problems in QDA. QDA.scalings_ " "contains values <= 0.0"
            )
        return self

    def predict(self, X):
        if self.estimator is None:
            raise NotImplementedError()
        return self.estimator.predict(X)

    def predict_proba(self, X):
        if self.estimator is None:
            raise NotImplementedError()

        df = self.estimator.predict_proba(X)
        return softmax(df)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "QDA",
            "name": "Quadratic Discriminant Analysis",
            "handles_regression": False,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": False,
            "is_deterministic": True,
            "input": (DENSE, UNSIGNED_DATA),
            "output": (PREDICTIONS,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        reg_param = UniformFloatHyperparameter("reg_param", 0.0, 1.0, default_value=0.0)
        cs = ConfigurationSpace()
        cs.add_hyperparameter(reg_param)
        return cs



import numpy as np
import scipy.sparse

from autosklearn.pipeline.components.feature_preprocessing.random_trees_embedding import (  # noqa: E501
    RandomTreesEmbedding,
)
from autosklearn.pipeline.util import _test_preprocessing, get_dataset

import unittest


class RandomTreesEmbeddingComponentTest(unittest.TestCase):
    def test_default_configuration(self):
        transformation, original = _test_preprocessing(RandomTreesEmbedding)
        self.assertEqual(transformation.shape[0], original.shape[0])
        self.assertEqual(transformation.shape[1], 218)
        self.assertIsInstance(original, np.ndarray)
        self.assertTrue(scipy.sparse.issparse(transformation))
        self.assertTrue(all(transformation.data == 1))

    @unittest.skip("Right now, the RTE returns a float64 array!")
    def test_preprocessing_dtype(self):
        # Dense
        # np.float32
        X_train, Y_train, X_test, Y_test = get_dataset("iris")
        self.assertEqual(X_train.dtype, np.float32)

        configuration_space = RandomTreesEmbedding.get_hyperparameter_search_space()
        default = configuration_space.get_default_configuration()
        preprocessor = RandomTreesEmbedding(
            random_state=1, **{hp_name: default[hp_name] for hp_name in default}
        )
        preprocessor.fit(X_train)
        Xt = preprocessor.transform(X_train)

        self.assertEqual(Xt.dtype, np.float32)

        # np.float64
        X_train, Y_train, X_test, Y_test = get_dataset("iris")
        X_train = X_train.astype(np.float64)
        configuration_space = RandomTreesEmbedding.get_hyperparameter_search_space()
        default = configuration_space.get_default_configuration()
        preprocessor = RandomTreesEmbedding(
            random_state=1, **{hp_name: default[hp_name] for hp_name in default}
        )
        preprocessor.fit(X_train, Y_train)
        Xt = preprocessor.transform(X_train)
        self.assertEqual(Xt.dtype, np.float64)



from autosklearn.ensemble_building.builder import EnsembleBuilder
from autosklearn.ensemble_building.manager import EnsembleBuilderManager
from autosklearn.ensemble_building.run import Run

__all__ = ["EnsembleBuilder", "EnsembleBuilderManager", "Run"]



from __future__ import annotations

import glob
import os
import pickle
import re
import shutil
import tempfile
import time
import uuid
import warnings
from typing import Dict, List, Optional, Tuple, TypeVar, Union, cast
from pathlib import Path

import numpy as np

import pandas as pd

import scipy.sparse

from sklearn.pipeline import Pipeline


from .logging_ import PicklableClientLogger, get_named_client_logger
from ..ensemble_building.abstract_ensemble import AbstractEnsemble


__all__ = ["Backend"]


DATAMANAGER_TYPE = TypeVar("DATAMANAGER_TYPE")
PIPELINE_IDENTIFIER_TYPE = Tuple[int, int, float]


def create(
    temporary_directory: str,
    output_directory: Optional[str],
    prefix: str,
    delete_tmp_folder_after_terminate: bool = True,
    delete_output_folder_after_terminate: bool = True,
) -> "Backend":
    context = BackendContext(
        temporary_directory,
        output_directory,
        delete_tmp_folder_after_terminate,
        delete_output_folder_after_terminate,
        prefix=prefix,
    )
    backend = Backend(context, prefix)

    return backend


def get_randomized_directory_name(
    prefix: str,
    temporary_directory: Optional[str] = None,
) -> str:
    uuid_str = str(uuid.uuid1(clock_seq=os.getpid()))

    temporary_directory = (
        temporary_directory
        if temporary_directory
        else os.path.join(
            tempfile.gettempdir(),
            "{}_tmp_{}".format(
                prefix,
                uuid_str,
            ),
        )
    )

    return temporary_directory


class BackendContext(object):
    def __init__(
        self,
        temporary_directory: str,
        output_directory: Optional[str],
        delete_tmp_folder_after_terminate: bool,
        delete_output_folder_after_terminate: bool,
        prefix: str,
    ):

        # Check that the names of tmp_dir and output_dir is not the same.
        if temporary_directory == output_directory and temporary_directory is not None:
            raise ValueError("The temporary and the output directory " "must be different.")

        self.delete_tmp_folder_after_terminate = delete_tmp_folder_after_terminate
        self.delete_output_folder_after_terminate = delete_output_folder_after_terminate
        # attributes to check that directories were properly created
        self._tmp_dir_created = False
        self._output_dir_created = False
        self._prefix = prefix

        self._temporary_directory = get_randomized_directory_name(
            temporary_directory=temporary_directory,
            prefix=self._prefix,
        )
        self._output_directory = output_directory
        # Logging happens via PicklableClientLogger
        # For this reason we need a port to communicate with the server
        # When the backend is created, this port is not available
        # When the port is available in the main process, we
        # call the setup_logger with this port and update self.logger
        self._logger = None  # type: Optional[PicklableClientLogger]
        self.create_directories()

    def setup_logger(self, port: int) -> None:
        self._logger = get_named_client_logger(
            name=__name__,
            port=port,
        )

    @property
    def output_directory(self) -> Optional[str]:
        if self._output_directory is not None:
            # make sure that tilde does not appear on the path.
            return os.path.expanduser(os.path.expandvars(self._output_directory))
        else:
            return None

    @property
    def temporary_directory(self) -> str:
        # make sure that tilde does not appear on the path.
        return os.path.expanduser(os.path.expandvars(self._temporary_directory))

    def create_directories(self, exist_ok: bool = False) -> None:
        # No Exception is raised if self.temporary_directory already exists.
        # This allows to continue the search, also, an error will be raised if
        # save_start_time is called which ensures two searches are not performed
        os.makedirs(self.temporary_directory, exist_ok=exist_ok)
        self._tmp_dir_created = True

        if self.output_directory is not None:
            os.makedirs(self.output_directory, exist_ok=exist_ok)
            self._output_dir_created = True

    def delete_directories(self, force: bool = True) -> None:
        if self.output_directory and (self.delete_output_folder_after_terminate or force):
            if self._output_dir_created is False:
                raise ValueError(
                    "Failed to delete output dir: %s"
                    "Please make sure that the specified output dir did not "
                    "previously exist." % self.output_directory
                )
            try:
                shutil.rmtree(self.output_directory)
            except Exception:
                try:
                    if self._logger is not None:
                        self._logger.warning(
                            "Could not delete output dir: %s" % self.output_directory
                        )
                    else:
                        print("Could not delete output dir: %s" % self.output_directory)
                except Exception:
                    print("Could not delete output dir: %s" % self.output_directory)

        if self.delete_tmp_folder_after_terminate or force:
            if self._tmp_dir_created is False:
                raise ValueError(
                    f"Failed to delete tmp dir {self.temporary_directory}."
                    "Please make sure that the specified tmp dir did not "
                    "previously exist."
                )
            try:
                shutil.rmtree(self.temporary_directory)
            except Exception:
                try:
                    if self._logger is not None:
                        self._logger.warning(
                            "Could not delete tmp dir: %s" % self.temporary_directory
                        )
                    else:
                        print("Could not delete tmp dir: %s" % self.temporary_directory)
                except Exception:
                    print("Could not delete tmp dir: %s" % self.temporary_directory)


class Backend(object):
    """Utility class to load and save all objects to be persisted.

    These are:
    * start time
    * true targets of the ensemble
    """

    def __init__(self, context: BackendContext, prefix: str):
        # When the backend is created, this port is not available
        # When the port is available in the main process, we
        # call the setup_logger with this port and update self.logger
        self.logger = None  # type: Optional[PicklableClientLogger]
        self.context = context
        self.prefix = prefix
        # Track the number of configurations launched
        # num_run == 1 means a dummy estimator run
        self.active_num_run = 1

        # Create the temporary directory if it does not yet exist
        try:
            os.makedirs(self.temporary_directory)
        except Exception:
            pass
        # This does not have to exist or be specified
        if self.output_directory is not None:
            if not os.path.exists(self.output_directory):
                raise ValueError("Output directory %s does not exist." % self.output_directory)

        self.internals_directory = os.path.join(self.temporary_directory, f".{self.prefix}")
        self._make_internals_directory()

    def setup_logger(self, port: int) -> None:
        self.logger = get_named_client_logger(
            name=__name__,
            port=port,
        )
        self.context.setup_logger(port)

    @property
    def output_directory(self) -> Optional[str]:
        return self.context.output_directory

    @property
    def temporary_directory(self) -> str:
        return self.context.temporary_directory

    def _make_internals_directory(self) -> None:
        try:
            os.makedirs(self.internals_directory)
        except Exception as e:
            if self.logger is not None:
                self.logger.debug("_make_internals_directory: %s" % e)
        try:
            os.makedirs(self.get_runs_directory())
        except Exception as e:
            if self.logger is not None:
                self.logger.debug("_make_internals_directory: %s" % e)

    def _get_start_time_filename(self, seed: Union[str, int]) -> str:
        if isinstance(seed, str):
            seed = int(seed)
        return os.path.join(self.internals_directory, "start_time_%d" % seed)

    def save_start_time(self, seed: str) -> str:
        self._make_internals_directory()
        start_time = time.time()

        filepath = self._get_start_time_filename(seed)

        if not isinstance(start_time, float):
            raise ValueError("Start time must be a float, but is %s." % type(start_time))

        if os.path.exists(filepath):
            raise ValueError(
                f"{filepath} already exist. Different seeds should be provided for different jobs."
            )

        with tempfile.NamedTemporaryFile("w", dir=os.path.dirname(filepath), delete=False) as fh:
            fh.write(str(start_time))
            tempname = fh.name
        os.rename(tempname, filepath)

        return filepath

    def load_start_time(self, seed: int) -> float:
        with open(self._get_start_time_filename(seed), "r") as fh:
            start_time = float(fh.read())
        return start_time

    def get_smac_output_directory(self) -> str:
        return os.path.join(self.temporary_directory, "smac3-output")

    def get_smac_output_directory_for_run(self, seed: int) -> str:
        return os.path.join(self.temporary_directory, "smac3-output", "run_%d" % seed)

    def _get_targets_ensemble_filename(self, end: str | None = None) -> str:
        dir = Path(self.internals_directory)
        stem = "true_targets_ensemble"
        existing = [p for p in dir.iterdir() if stem in p.name]

        # Sanity check to make sure we have one or None
        assert len(existing) in [0, 1]

        if not any(existing):
            end = "npy"
        else:
            end = existing[0].name.split('.')[-1]

        return os.path.join(self.internals_directory, f"{stem}.{end}")

    def _get_input_ensemble_filename(self, end: str | None = None) -> str:
        dir = Path(self.internals_directory)
        stem = "true_input_ensemble"
        existing = [p for p in dir.iterdir() if stem in p.name]

        # Sanity check to make sure we have one or None
        assert len(existing) in [0, 1]

        if not any(existing):
            end = "npy"
        else:
            end = existing[0].name.split('.')[-1]

        return os.path.join(self.internals_directory, f"{stem}.{end}")

    def save_additional_data(
        self,
        data: Union[np.ndarray, pd.DataFrame, scipy.sparse.spmatrix],
        what: str,
        overwrite: bool = False,
    ) -> str:
        self._make_internals_directory()
        if isinstance(data, np.ndarray):
            end = "npy"
        elif isinstance(data, scipy.sparse.spmatrix):
            end = "npz"
        elif isinstance(data, pd.DataFrame):
            end = "pd"
        else:
            raise ValueError(
                "Targets must be of type np.ndarray, pd.Dataframe or"
                " scipy.sparse.spmatrix but is %s" % type(data)
            )

        if what == "targets_ensemble":
            filepath = self._get_targets_ensemble_filename(end=end)
        elif what == "input_ensemble":
            filepath = self._get_input_ensemble_filename(end=end)
        else:
            raise ValueError(f"Unknown data type {what}")

        # Only store data if it does not exist yet
        if not overwrite and os.path.isfile(filepath):
            return filepath

        tempname = self._save_array(data=data, filepath=filepath)
        os.rename(tempname, filepath)

        return filepath

    @staticmethod
    def _save_array(
        data: Union[np.ndarray, pd.DataFrame, scipy.sparse.spmatrix], filepath: str
    ) -> str:
        if isinstance(data, np.ndarray):
            with tempfile.NamedTemporaryFile(
                "wb", dir=os.path.dirname(filepath), delete=False
            ) as fh_w:
                np.save(fh_w, data.astype(np.float32))
        elif isinstance(data, scipy.sparse.spmatrix):
            with tempfile.NamedTemporaryFile(
                "wb", dir=os.path.dirname(filepath), delete=False
            ) as fh_w:
                scipy.sparse.save_npz(fh_w, data)
        elif isinstance(data, pd.DataFrame):
            with tempfile.NamedTemporaryFile(
                "wb", dir=os.path.dirname(filepath), delete=False
            ) as fh_w:
                data.to_pickle(fh_w)
        return fh_w.name

    @staticmethod
    def _load_array(filepath: str) -> np.array:
        end = filepath.split(".")[-1]
        if end == "npy":
            targets = np.load(filepath, allow_pickle=True)
        elif end == "npz":
            targets = scipy.sparse.load_npz(filepath)
        elif end == "pd":
            targets = pd.read_pickle(filepath)
        else:
            raise ValueError(f"Unknown file type {end} in {filepath}")

        return targets

    def load_targets_ensemble(self) -> np.ndarray:
        return self._load_array(filepath=self._get_targets_ensemble_filename())

    def load_input_ensemble(self) -> np.ndarray:
        return self._load_array(filepath=self._get_input_ensemble_filename())

    def _get_datamanager_pickle_filename(self) -> str:
        return os.path.join(self.internals_directory, "datamanager.pkl")

    def save_datamanager(self, datamanager: DATAMANAGER_TYPE) -> str:
        self._make_internals_directory()
        filepath = self._get_datamanager_pickle_filename()

        with tempfile.NamedTemporaryFile("wb", dir=os.path.dirname(filepath), delete=False) as fh:
            pickle.dump(datamanager, fh, -1)
            tempname = fh.name
        os.rename(tempname, filepath)

        return filepath

    def load_datamanager(self) -> DATAMANAGER_TYPE:
        filepath = self._get_datamanager_pickle_filename()
        with open(filepath, "rb") as fh:
            return cast(DATAMANAGER_TYPE, pickle.load(fh))

    def get_runs_directory(self) -> str:
        return os.path.join(self.internals_directory, "runs")

    def get_numrun_directory(self, seed: int, num_run: int, budget: float) -> str:
        return os.path.join(self.internals_directory, "runs", "%d_%d_%s" % (seed, num_run, budget))

    def get_next_num_run(self, peek: bool = False) -> int:
        """
        Every pipeline that is fitted by the estimator is stored with an
        identifier called num_run. A dummy classifier will always have a num_run
        equal to 1, and all other new configurations that are explored will
        have a sequentially increasing identifier.

        This method returns the next num_run a configuration should take.

        Parameters
        ----------
        peek: bool
            By default, the next num_rum will be returned, i.e. self.active_num_run + 1
            Yet, if this bool parameter is equal to True, the value of the current
            num_run is provided, i.e, self.active_num_run.
            In other words, peek allows to get the current maximum identifier
            of a configuration.

        Returns
        -------
        num_run: int
            An unique identifier for a configuration
        """

        # If there are other num_runs, their name would be runs/<seed>_<num_run>_<budget>
        other_num_runs = [
            int(os.path.basename(run_dir).split("_")[1])
            for run_dir in glob.glob(os.path.join(self.internals_directory, "runs", "*"))
            if self._is_run_dir(os.path.basename(run_dir))
        ]
        if len(other_num_runs) > 0:
            # We track the number of runs from two forefronts:
            # The physically available num_runs (which might be deleted or a crash could happen)
            # From a internally kept attribute. The later should be sufficient, but we
            # want to be robust against multiple backend copies on different workers
            self.active_num_run = max([self.active_num_run] + other_num_runs)

        # We are interested in the next run id
        if not peek:
            self.active_num_run += 1
        return self.active_num_run

    @staticmethod
    def _is_run_dir(run_dir: str) -> bool:
        """
        Run directories are stored in the format <seed>_<num_run>_<budget>.

        Parameters
        ----------
        run_dir: str
            string containing the base name of the run directory

        Returns
        -------
        _: bool
            whether the provided run directory matches the run_dir_pattern
            signifying that it is a run directory
        """
        run_dir_pattern = r"\d+_\d+_\d+"
        return bool(re.match(run_dir_pattern, run_dir))

    def get_model_filename(self, seed: int, idx: int, budget: float) -> str:
        return "%s.%s.%s.model" % (seed, idx, budget)

    def get_cv_model_filename(self, seed: int, idx: int, budget: float) -> str:
        return "%s.%s.%s.cv_model" % (seed, idx, budget)

    def list_all_models(self, seed: int) -> List[str]:
        runs_directory = self.get_runs_directory()
        model_files = glob.glob(
            os.path.join(glob.escape(runs_directory), "%d_*" % seed, "%s.*.*.model" % seed)
        )
        return model_files

    def load_models_by_identifiers(
        self, identifiers: List[PIPELINE_IDENTIFIER_TYPE]
    ) -> Dict[PIPELINE_IDENTIFIER_TYPE, Pipeline]:
        models = {}

        for identifier in identifiers:
            seed, idx, budget = identifier
            models[identifier] = self.load_model_by_seed_and_id_and_budget(seed, idx, budget)

        return models

    def load_model_by_seed_and_id_and_budget(self, seed: int, idx: int, budget: float) -> Pipeline:
        model_directory = self.get_numrun_directory(seed, idx, budget)

        model_file_name = "%s.%s.%s.model" % (seed, idx, budget)
        model_file_path = os.path.join(model_directory, model_file_name)
        with open(model_file_path, "rb") as fh:
            return pickle.load(fh)

    def load_cv_models_by_identifiers(
        self, identifiers: List[PIPELINE_IDENTIFIER_TYPE]
    ) -> Dict[PIPELINE_IDENTIFIER_TYPE, Pipeline]:
        models = {}

        for identifier in identifiers:
            seed, idx, budget = identifier
            models[identifier] = self.load_cv_model_by_seed_and_id_and_budget(seed, idx, budget)

        return models

    def load_cv_model_by_seed_and_id_and_budget(
        self, seed: int, idx: int, budget: float
    ) -> Pipeline:
        model_directory = self.get_numrun_directory(seed, idx, budget)

        model_file_name = "%s.%s.%s.cv_model" % (seed, idx, budget)
        model_file_path = os.path.join(model_directory, model_file_name)
        with open(model_file_path, "rb") as fh:
            return pickle.load(fh)

    def save_numrun_to_dir(
        self,
        seed: int,
        idx: int,
        budget: float,
        model: Optional[Pipeline],
        cv_model: Optional[Pipeline],
        ensemble_predictions: Optional[np.ndarray],
        valid_predictions: Optional[np.ndarray],
        test_predictions: Optional[np.ndarray],
    ) -> None:
        runs_directory = self.get_runs_directory()
        tmpdir = tempfile.mkdtemp(dir=runs_directory)
        if model is not None:
            file_path = os.path.join(tmpdir, self.get_model_filename(seed, idx, budget))
            with open(file_path, "wb") as fh:
                pickle.dump(model, fh, -1)

        if cv_model is not None:
            file_path = os.path.join(tmpdir, self.get_cv_model_filename(seed, idx, budget))
            with open(file_path, "wb") as fh:
                pickle.dump(cv_model, fh, -1)

        for preds, subset in (
            (ensemble_predictions, "ensemble"),
            (valid_predictions, "valid"),
            (test_predictions, "test"),
        ):
            if preds is not None:
                file_path = os.path.join(
                    tmpdir, self.get_prediction_filename(subset, seed, idx, budget)
                )
                with open(file_path, "wb") as fh:
                    pickle.dump(preds.astype(np.float32), fh, -1)
        try:
            os.rename(tmpdir, self.get_numrun_directory(seed, idx, budget))
        except OSError:
            if os.path.exists(self.get_numrun_directory(seed, idx, budget)):
                os.rename(
                    self.get_numrun_directory(seed, idx, budget),
                    os.path.join(runs_directory, tmpdir + ".old"),
                )
                os.rename(tmpdir, self.get_numrun_directory(seed, idx, budget))
                shutil.rmtree(os.path.join(runs_directory, tmpdir + ".old"))

    def get_ensemble_dir(self) -> str:
        return os.path.join(self.internals_directory, "ensembles")

    def load_ensemble(self, seed: int) -> Optional[AbstractEnsemble]:
        ensemble_dir = self.get_ensemble_dir()

        if not os.path.exists(ensemble_dir):
            if self.logger is not None:
                self.logger.warning("Directory %s does not exist" % ensemble_dir)
            else:
                warnings.warn("Directory %s does not exist" % ensemble_dir)
            return None

        if seed >= 0:
            indices_files = glob.glob(
                os.path.join(glob.escape(ensemble_dir), "%s.*.ensemble" % seed)
            )
            indices_files.sort()
        else:
            indices_files = os.listdir(ensemble_dir)
            indices_files = [os.path.join(ensemble_dir, f) for f in indices_files]
            indices_files.sort(key=lambda f: time.ctime(os.path.getmtime(f)))

        with open(indices_files[-1], "rb") as fh:
            ensemble_members_run_numbers = cast(AbstractEnsemble, pickle.load(fh))

        return ensemble_members_run_numbers

    def save_ensemble(self, ensemble: AbstractEnsemble, idx: int, seed: int) -> None:
        try:
            os.makedirs(self.get_ensemble_dir())
        except Exception:
            pass

        filepath = os.path.join(
            self.get_ensemble_dir(), "%s.%s.ensemble" % (str(seed), str(idx).zfill(10))
        )
        with tempfile.NamedTemporaryFile("wb", dir=os.path.dirname(filepath), delete=False) as fh:
            pickle.dump(ensemble, fh)
            tempname = fh.name
        os.rename(tempname, filepath)

    def get_prediction_filename(
        self, subset: str, automl_seed: Union[str, int], idx: int, budget: float
    ) -> str:
        return "predictions_%s_%s_%s_%s.npy" % (subset, automl_seed, idx, budget)

    def save_predictions_as_txt(
        self,
        predictions: np.ndarray,
        subset: str,
        idx: int,
        precision: int,
        prefix: Optional[str] = None,
    ) -> None:
        if not self.output_directory:
            return
        # Write prediction scores in prescribed format
        filepath = os.path.join(
            self.output_directory,
            ("%s_" % prefix if prefix else "") + "%s_%s.predict" % (subset, str(idx)),
        )

        format_string = "{:.%dg} " % precision
        with tempfile.NamedTemporaryFile(
            "w", dir=os.path.dirname(filepath), delete=False
        ) as output_file:
            for row in predictions:
                if not isinstance(row, np.ndarray) and not isinstance(row, list):
                    row = [row]
                for val in row:
                    output_file.write(format_string.format(float(val)))
                output_file.write("\n")
            tempname = output_file.name
        os.rename(tempname, filepath)

    def write_txt_file(self, filepath: str, data: str, name: str) -> None:
        with tempfile.NamedTemporaryFile("w", dir=os.path.dirname(filepath), delete=False) as fh:
            fh.write(data)
            tempname = fh.name
        os.rename(tempname, filepath)
        if self.logger is not None:
            self.logger.debug("Created %s file %s" % (name, filepath))



# -*- encoding: utf-8 -*-
import logging.handlers
import os
import shutil
import sys
import tempfile

import numpy as np
import sklearn.dummy
from smac.tae import StatusType

from autosklearn.automl_common.common.utils.backend import Backend, BackendContext
from autosklearn.evaluation.abstract_evaluator import AbstractEvaluator
from autosklearn.metrics import accuracy
from autosklearn.pipeline.components.base import _addons

import unittest
import unittest.mock

this_directory = os.path.dirname(__file__)
sys.path.append(this_directory)
from evaluation_util import get_multiclass_classification_datamanager  # noqa E402


class AbstractEvaluatorTest(unittest.TestCase):
    _multiprocess_can_split_ = True

    def setUp(self):
        """
        Creates a backend mock
        """
        self.ev_path = os.path.join(this_directory, ".tmp_evaluations")
        if not os.path.exists(self.ev_path):
            os.mkdir(self.ev_path)
        dummy_model_files = [os.path.join(self.ev_path, str(n)) for n in range(100)]
        dummy_pred_files = [os.path.join(self.ev_path, str(n)) for n in range(100, 200)]

        backend_mock = unittest.mock.Mock()
        backend_mock.get_model_dir.return_value = self.ev_path
        backend_mock.get_model_path.side_effect = dummy_model_files
        backend_mock.get_prediction_output_path.side_effect = dummy_pred_files
        D = get_multiclass_classification_datamanager()
        backend_mock.load_datamanager.return_value = D
        backend_mock.temporary_directory = tempfile.gettempdir()
        self.backend_mock = backend_mock

        self.port = logging.handlers.DEFAULT_TCP_LOGGING_PORT

        self.working_directory = os.path.join(this_directory, ".tmp_%s" % self.id())

    def tearDown(self):
        if os.path.exists(self.ev_path):
            try:
                os.rmdir(self.ev_path)
            except:  # noqa E722
                pass

    def test_finish_up_model_predicts_NaN(self):
        """Tests by handing in predictions which contain NaNs"""
        rs = np.random.RandomState(1)

        queue_mock = unittest.mock.Mock()
        ae = AbstractEvaluator(
            backend=self.backend_mock,
            port=self.port,
            output_y_hat_optimization=False,
            queue=queue_mock,
            metrics=[accuracy],
            additional_components=dict(),
        )
        ae.Y_optimization = rs.rand(33, 3)
        predictions_ensemble = rs.rand(33, 3)
        predictions_test = rs.rand(25, 3)

        # NaNs in prediction ensemble
        predictions_ensemble[5, 2] = np.NaN
        _, loss, _, additional_run_info = ae.finish_up(
            loss=0.1,
            train_loss=0.1,
            opt_pred=predictions_ensemble,
            test_pred=predictions_test,
            additional_run_info=None,
            final_call=True,
            file_output=True,
            status=StatusType.SUCCESS,
        )
        self.assertEqual(loss, 1.0)
        self.assertEqual(
            additional_run_info,
            {"error": "Model predictions for optimization set contains NaNs."},
        )

        predictions_ensemble = rs.rand(33, 3)
        predictions_test[5, 2] = np.NaN
        _, loss, _, additional_run_info = ae.finish_up(
            loss=0.1,
            train_loss=0.1,
            opt_pred=predictions_ensemble,
            test_pred=predictions_test,
            additional_run_info=None,
            final_call=True,
            file_output=True,
            status=StatusType.SUCCESS,
        )
        self.assertEqual(loss, 1.0)
        self.assertEqual(
            additional_run_info,
            {"error": "Model predictions for test set contains NaNs."},
        )
        self.assertEqual(self.backend_mock.save_predictions_as_npy.call_count, 0)

    def test_disable_file_output(self):
        queue_mock = unittest.mock.Mock()

        rs = np.random.RandomState(1)

        ae = AbstractEvaluator(
            backend=self.backend_mock,
            queue=queue_mock,
            disable_file_output=True,
            metrics=[accuracy],
            port=self.port,
            additional_components=dict(),
        )

        predictions_ensemble = rs.rand(33, 3)
        predictions_test = rs.rand(25, 3)

        loss_, additional_run_info_ = ae.file_output(
            predictions_ensemble,
            predictions_test,
        )

        self.assertIsNone(loss_)
        self.assertEqual(additional_run_info_, {})
        # This function is never called as there is a return before
        self.assertEqual(self.backend_mock.save_numrun_to_dir.call_count, 0)

        for call_count, disable in enumerate(["model", "cv_model"], start=1):
            ae = AbstractEvaluator(
                backend=self.backend_mock,
                output_y_hat_optimization=False,
                queue=queue_mock,
                disable_file_output=[disable],
                metrics=[accuracy],
                port=self.port,
                additional_components=dict(),
            )
            ae.Y_optimization = predictions_ensemble
            ae.model = unittest.mock.Mock()
            ae.models = [unittest.mock.Mock()]

            loss_, additional_run_info_ = ae.file_output(
                predictions_ensemble,
                predictions_test,
            )

            self.assertIsNone(loss_)
            self.assertEqual(additional_run_info_, {})
            self.assertEqual(
                self.backend_mock.save_numrun_to_dir.call_count, call_count
            )
            if disable == "model":
                self.assertIsNone(
                    self.backend_mock.save_numrun_to_dir.call_args_list[-1][1]["model"]
                )
                self.assertIsNotNone(
                    self.backend_mock.save_numrun_to_dir.call_args_list[-1][1][
                        "cv_model"
                    ]
                )
            else:
                self.assertIsNotNone(
                    self.backend_mock.save_numrun_to_dir.call_args_list[-1][1]["model"]
                )
                self.assertIsNone(
                    self.backend_mock.save_numrun_to_dir.call_args_list[-1][1][
                        "cv_model"
                    ]
                )
            self.assertIsNotNone(
                self.backend_mock.save_numrun_to_dir.call_args_list[-1][1][
                    "ensemble_predictions"
                ]
            )
            self.assertIsNotNone(
                self.backend_mock.save_numrun_to_dir.call_args_list[-1][1][
                    "test_predictions"
                ]
            )

        ae = AbstractEvaluator(
            backend=self.backend_mock,
            output_y_hat_optimization=False,
            queue=queue_mock,
            metrics=[accuracy],
            disable_file_output=["y_optimization"],
            port=self.port,
            additional_components=dict(),
        )
        ae.Y_optimization = predictions_ensemble
        ae.model = "model"
        ae.models = [unittest.mock.Mock()]

        loss_, additional_run_info_ = ae.file_output(
            predictions_ensemble,
            predictions_test,
        )

        self.assertIsNone(loss_)
        self.assertEqual(additional_run_info_, {})

        self.assertIsNone(
            self.backend_mock.save_numrun_to_dir.call_args_list[-1][1][
                "ensemble_predictions"
            ]
        )
        self.assertIsNotNone(
            self.backend_mock.save_numrun_to_dir.call_args_list[-1][1][
                "test_predictions"
            ]
        )

    def test_file_output(self):
        shutil.rmtree(self.working_directory, ignore_errors=True)
        os.mkdir(self.working_directory)

        queue_mock = unittest.mock.Mock()

        context = BackendContext(
            temporary_directory=os.path.join(self.working_directory, "tmp"),
            output_directory=os.path.join(self.working_directory, "tmp_output"),
            delete_tmp_folder_after_terminate=True,
            delete_output_folder_after_terminate=True,
            prefix="auto-sklearn",
        )
        with unittest.mock.patch.object(
            Backend, "load_datamanager"
        ) as load_datamanager_mock:
            load_datamanager_mock.return_value = (
                get_multiclass_classification_datamanager()
            )

            backend = Backend(context, prefix="auto-sklearn")

            ae = AbstractEvaluator(
                backend=backend,
                output_y_hat_optimization=False,
                queue=queue_mock,
                metrics=[accuracy],
                port=self.port,
                additional_components=dict(),
            )
            ae.model = sklearn.dummy.DummyClassifier()

            rs = np.random.RandomState(1)
            ae.Y_optimization = rs.rand(33, 3)
            predictions_ensemble = rs.rand(33, 3)
            predictions_test = rs.rand(25, 3)

            ae.file_output(
                Y_optimization_pred=predictions_ensemble,
                Y_test_pred=predictions_test,
            )

            self.assertTrue(
                os.path.exists(
                    os.path.join(
                        self.working_directory,
                        "tmp",
                        ".auto-sklearn",
                        "runs",
                        "1_0_None",
                    )
                )
            )

            shutil.rmtree(self.working_directory, ignore_errors=True)

    def test_add_additional_components(self):
        shutil.rmtree(self.working_directory, ignore_errors=True)
        os.mkdir(self.working_directory)

        queue_mock = unittest.mock.Mock()

        context = BackendContext(
            temporary_directory=os.path.join(self.working_directory, "tmp"),
            output_directory=os.path.join(self.working_directory, "tmp_output"),
            delete_tmp_folder_after_terminate=True,
            delete_output_folder_after_terminate=True,
            prefix="auto-sklearn",
        )
        with unittest.mock.patch.object(
            Backend, "load_datamanager"
        ) as load_datamanager_mock:
            load_datamanager_mock.return_value = (
                get_multiclass_classification_datamanager()
            )
            backend = Backend(context, prefix="auto-sklearn")

            with unittest.mock.patch.object(
                _addons["classification"], "add_component"
            ) as _:

                # If the components in the argument `additional_components` are an
                # empty dict there is no call to `add_component`,
                # if there's something in it, `add_component is called (2nd case)
                for fixture, case in ((0, dict()), (1, dict(abc="def"))):

                    thirdparty_components_patch = unittest.mock.Mock()
                    thirdparty_components_patch.components = case
                    additional_components = dict(
                        classification=thirdparty_components_patch
                    )
                    AbstractEvaluator(
                        backend=backend,
                        output_y_hat_optimization=False,
                        queue=queue_mock,
                        metrics=[accuracy],
                        port=self.port,
                        additional_components=additional_components,
                    )
                    self.assertEqual(
                        _addons["classification"].add_component.call_count, fixture
                    )



import numpy as np
import scipy.sparse
import sklearn.preprocessing

from autosklearn.pipeline.components.feature_preprocessing.select_percentile_classification import (  # noqa: E501
    SelectPercentileClassification,
)
from autosklearn.pipeline.util import _test_preprocessing, get_dataset

import unittest


class SelectPercentileClassificationTest(unittest.TestCase):
    def test_default_configuration(self):
        transformation, original = _test_preprocessing(SelectPercentileClassification)
        self.assertEqual(transformation.shape[0], original.shape[0])
        self.assertEqual(transformation.shape[1], int(original.shape[1] / 2))
        self.assertFalse((transformation == 0).all())

        transformation, original = _test_preprocessing(
            SelectPercentileClassification,
            make_sparse=True,
        )
        self.assertTrue(scipy.sparse.issparse(transformation))
        self.assertEqual(transformation.shape[0], original.shape[0])
        self.assertEqual(transformation.shape[1], int(original.shape[1] / 2))

        # Custon preprocessing test to check if clipping to zero works
        X_train, Y_train, X_test, Y_test = get_dataset(dataset="digits")
        original_X_train = X_train.copy()
        ss = sklearn.preprocessing.StandardScaler()
        X_train = ss.fit_transform(X_train)
        configuration_space = (
            SelectPercentileClassification.get_hyperparameter_search_space()
        )
        default = configuration_space.get_default_configuration()

        preprocessor = SelectPercentileClassification(
            random_state=1,
            **{
                hp_name: default[hp_name]
                for hp_name in default
                if default[hp_name] is not None
            },
        )

        transformer = preprocessor.fit(X_train, Y_train)
        transformation, original = transformer.transform(X_train), original_X_train
        self.assertEqual(transformation.shape[0], original.shape[0])
        self.assertEqual(transformation.shape[1], int(original.shape[1] / 2))

    def test_preprocessing_dtype(self):
        # Dense
        # np.float32
        X_train, Y_train, X_test, Y_test = get_dataset("iris")
        self.assertEqual(X_train.dtype, np.float32)

        configuration_space = (
            SelectPercentileClassification.get_hyperparameter_search_space()
        )
        default = configuration_space.get_default_configuration()
        preprocessor = SelectPercentileClassification(
            random_state=1, **{hp_name: default[hp_name] for hp_name in default}
        )
        preprocessor.fit(X_train, Y_train)
        Xt = preprocessor.transform(X_train)
        self.assertEqual(Xt.dtype, np.float32)

        # np.float64
        X_train, Y_train, X_test, Y_test = get_dataset("iris")
        X_train = X_train.astype(np.float64)
        configuration_space = (
            SelectPercentileClassification.get_hyperparameter_search_space()
        )
        default = configuration_space.get_default_configuration()
        preprocessor = SelectPercentileClassification(
            random_state=1, **{hp_name: default[hp_name] for hp_name in default}
        )
        preprocessor.fit(X_train, Y_train)
        Xt = preprocessor.transform(X_train)
        self.assertEqual(Xt.dtype, np.float64)

        # Sparse
        # np.float32
        X_train, Y_train, X_test, Y_test = get_dataset("iris", make_sparse=True)
        self.assertEqual(X_train.dtype, np.float32)
        configuration_space = (
            SelectPercentileClassification.get_hyperparameter_search_space()
        )
        default = configuration_space.get_default_configuration()
        preprocessor = SelectPercentileClassification(
            random_state=1, **{hp_name: default[hp_name] for hp_name in default}
        )
        preprocessor.fit(X_train, Y_train)
        Xt = preprocessor.transform(X_train)
        self.assertEqual(Xt.dtype, np.float32)

        # np.float64
        X_train, Y_train, X_test, Y_test = get_dataset("iris", make_sparse=True)
        X_train = X_train.astype(np.float64)
        configuration_space = (
            SelectPercentileClassification.get_hyperparameter_search_space()
        )
        default = configuration_space.get_default_configuration()
        preprocessor = SelectPercentileClassification(
            random_state=1, **{hp_name: default[hp_name] for hp_name in default}
        )
        preprocessor.fit(X_train, Y_train)
        Xt = preprocessor.transform(X_train)
        self.assertEqual(Xt.dtype, np.float64)



from typing import Dict, Optional, Tuple, Union

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import CategoricalHyperparameter

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA


class NumericalImputation(AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self,
        strategy: str = "mean",
        random_state: Optional[Union[int, np.random.RandomState]] = None,
    ) -> None:
        self.strategy = strategy
        self.random_state = random_state

    def fit(
        self, X: PIPELINE_DATA_DTYPE, y: Optional[PIPELINE_DATA_DTYPE] = None
    ) -> "NumericalImputation":
        import sklearn.impute

        self.preprocessor = sklearn.impute.SimpleImputer(
            strategy=self.strategy, copy=False
        )
        self.preprocessor.fit(X)
        return self

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        if self.preprocessor is None:
            raise NotImplementedError()
        return self.preprocessor.transform(X)

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "NumericalImputation",
            "name": "Numerical Imputation",
            "handles_missing_values": True,
            "handles_nominal_values": True,
            "handles_numerical_features": True,
            "prefers_data_scaled": False,
            "prefers_data_normalized": False,
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "is_deterministic": True,
            # TODO find out if this is right!
            "handles_sparse": True,
            "handles_dense": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
            "preferred_dtype": None,
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> ConfigurationSpace:
        # TODO add replace by zero!
        strategy = CategoricalHyperparameter(
            "strategy", ["mean", "median", "most_frequent"], default_value="mean"
        )
        cs = ConfigurationSpace()
        cs.add_hyperparameter(strategy)
        return cs



from autosklearn.pipeline.components.data_preprocessing.feature_type import (
    FeatTypeSplit,
)

import unittest


class PreprocessingPipelineFeatTypeTest(unittest.TestCase):

    num_numerical = 6
    num_categorical = 3
    num_text = 11

    def test_single_type(self):
        DPP = FeatTypeSplit(feat_type={"A": "numerical"})
        cs = DPP.get_hyperparameter_search_space(
            feat_type={"A": "numerical"},
            dataset_properties={
                "task": 1,
                "sparse": False,
                "multilabel": False,
                "multiclass": False,
                "target_type": "classification",
                "signed": False,
            },
        )
        for key in cs.get_hyperparameters_dict().keys():
            self.assertNotIn("text", key.split(":")[0])
            self.assertNotIn("categorical", key.split(":")[0])
        self.assertEqual(len(cs), self.num_numerical)

        DPP = FeatTypeSplit(feat_type={"A": "categorical"})
        cs = DPP.get_hyperparameter_search_space(
            feat_type={"A": "categorical"},
            dataset_properties={
                "task": 1,
                "sparse": False,
                "multilabel": False,
                "multiclass": False,
                "target_type": "classification",
                "signed": False,
            },
        )
        for key in cs.get_hyperparameters_dict().keys():
            self.assertNotIn("text", key.split(":")[0])
            self.assertNotIn("numerical", key.split(":")[0])
        self.assertEqual(len(cs), self.num_categorical)

        DPP = FeatTypeSplit(feat_type={"A": "string"})
        cs = DPP.get_hyperparameter_search_space(
            feat_type={"A": "string"},
            dataset_properties={
                "task": 1,
                "sparse": False,
                "multilabel": False,
                "multiclass": False,
                "target_type": "classification",
                "signed": False,
            },
        )
        for key in cs.get_hyperparameters_dict().keys():
            self.assertNotIn("numerical", key.split(":")[0])
            self.assertNotIn("categorical", key.split(":")[0])
        self.assertEqual(len(cs), self.num_text)

    def test_dual_type(self):
        DPP = FeatTypeSplit(feat_type={"A": "numerical", "B": "categorical"})
        cs = DPP.get_hyperparameter_search_space(
            feat_type={"A": "numerical", "B": "categorical"},
            dataset_properties={
                "task": 1,
                "sparse": False,
                "multilabel": False,
                "multiclass": False,
                "target_type": "classification",
                "signed": False,
            },
        )
        for key in cs.get_hyperparameters_dict().keys():
            self.assertNotIn("text", key.split(":")[0])
        self.assertEqual(len(cs), self.num_numerical + self.num_categorical)

        DPP = FeatTypeSplit(feat_type={"A": "categorical", "B": "string"})
        cs = DPP.get_hyperparameter_search_space(
            feat_type={"A": "categorical", "B": "string"},
            dataset_properties={
                "task": 1,
                "sparse": False,
                "multilabel": False,
                "multiclass": False,
                "target_type": "classification",
                "signed": False,
            },
        )
        for key in cs.get_hyperparameters_dict().keys():
            self.assertNotIn("numerical", key.split(":")[0])
        self.assertEqual(len(cs), self.num_categorical + self.num_text)

        DPP = FeatTypeSplit(feat_type={"A": "string", "B": "numerical"})
        cs = DPP.get_hyperparameter_search_space(
            feat_type={"A": "string", "B": "numerical"},
            dataset_properties={
                "task": 1,
                "sparse": False,
                "multilabel": False,
                "multiclass": False,
                "target_type": "classification",
                "signed": False,
            },
        )
        for key in cs.get_hyperparameters_dict().keys():
            self.assertNotIn("categorical", key.split(":")[0])
        self.assertEqual(len(cs), self.num_text + self.num_numerical)

    def test_triple_type(self):
        DPP = FeatTypeSplit(
            feat_type={"A": "numerical", "B": "categorical", "C": "string"}
        )
        cs = DPP.get_hyperparameter_search_space(
            feat_type={"A": "numerical", "B": "categorical", "C": "string"},
            dataset_properties={
                "task": 1,
                "sparse": False,
                "multilabel": False,
                "multiclass": False,
                "target_type": "classification",
                "signed": False,
            },
        )
        truth_table = [False] * 3
        for key in cs.get_hyperparameters_dict().keys():
            if "text" in key.split(":")[0]:
                truth_table[0] = True
            elif "categorical" in key.split(":")[0]:
                truth_table[1] = True
            elif "numerical" in key.split(":")[0]:
                truth_table[2] = True

        self.assertEqual(sum(truth_table), 3)
        self.assertEqual(
            len(cs), self.num_numerical + self.num_categorical + self.num_text
        )



from collections import OrderedDict

import numpy
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import CategoricalHyperparameter

import autosklearn.pipeline.create_searchspace_util
from autosklearn.pipeline.components.classification.lda import LDA
from autosklearn.pipeline.components.classification.liblinear_svc import LibLinear_SVC
from autosklearn.pipeline.components.feature_preprocessing.no_preprocessing import (
    NoPreprocessing,
)
from autosklearn.pipeline.components.feature_preprocessing.pca import PCA
from autosklearn.pipeline.components.feature_preprocessing.random_trees_embedding import (  # noqa: E501
    RandomTreesEmbedding,
)
from autosklearn.pipeline.components.feature_preprocessing.truncatedSVD import (
    TruncatedSVD,
)

import unittest


class TestCreateClassificationSearchspace(unittest.TestCase):
    _multiprocess_can_split_ = True

    def test_get_match_array_sparse_and_dense(self):
        # preproc is empty
        preprocessors = OrderedDict()
        preprocessors["pca"] = PCA
        classifiers = OrderedDict()
        classifiers["lda"] = LDA
        # Sparse + dense

        class Preprocessors(object):
            @classmethod
            def get_available_components(self, *args, **kwargs):
                return preprocessors

        class Classifiers(object):
            @classmethod
            def get_available_components(self, *args, **kwargs):
                return classifiers

        # Dense
        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, PCA), (1, LDA)), dataset_properties={"sparse": True}
        )
        self.assertEqual(numpy.sum(m), 0)

        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, PCA), (1, LDA)), dataset_properties={"sparse": False}
        )
        self.assertEqual(m, [[1]])

        # Sparse
        preprocessors["tSVD"] = TruncatedSVD
        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, LDA)), dataset_properties={"sparse": True}
        )
        self.assertEqual(m[0], [0])  # pca
        self.assertEqual(m[1], [1])  # svd

        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, LDA)),
            dataset_properties={"sparse": False},
        )
        self.assertEqual(m[0], [1])  # pca
        self.assertEqual(m[1], [0])  # svd

        preprocessors["none"] = NoPreprocessing
        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, LDA)), dataset_properties={"sparse": True}
        )
        self.assertEqual(m[0, :], [0])  # pca
        self.assertEqual(m[1, :], [1])  # tsvd
        self.assertEqual(m[2, :], [0])  # none

        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, LDA)),
            dataset_properties={"sparse": False},
        )
        self.assertEqual(m[0, :], [1])  # pca
        self.assertEqual(m[1, :], [0])  # tsvd
        self.assertEqual(m[2, :], [1])  # none

        classifiers["libsvm"] = LibLinear_SVC
        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, Classifiers)),
            dataset_properties={"sparse": False},
        )
        self.assertListEqual(list(m[0, :]), [1, 1])  # pca
        self.assertListEqual(list(m[1, :]), [0, 0])  # tsvd
        self.assertListEqual(list(m[2, :]), [1, 1])  # none

        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, Classifiers)),
            dataset_properties={"sparse": True},
        )
        self.assertListEqual(list(m[0, :]), [0, 0])  # pca
        self.assertListEqual(list(m[1, :]), [1, 1])  # tsvd
        self.assertListEqual(list(m[2, :]), [0, 1])  # none

        # Do fancy 3d stuff
        preprocessors["random_trees"] = RandomTreesEmbedding
        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, Preprocessors), (2, Classifiers)),
            dataset_properties={"sparse": False},
        )
        # PCA followed by truncated SVD is forbidden
        self.assertEqual(list(m[0].flatten()), [1, 1, 0, 0, 1, 1, 0, 1])
        # Truncated SVD is forbidden
        self.assertEqual(list(m[1].flatten()), [0, 0, 0, 0, 0, 0, 0, 0])
        # Truncated SVD is forbidden after no_preprocessing
        self.assertEqual(list(m[2].flatten()), [1, 1, 0, 0, 1, 1, 0, 1])
        # PCA is forbidden, truncatedSVD allowed after random trees embedding
        # lda only allowed after truncatedSVD
        self.assertEqual(list(m[3].flatten()), [0, 0, 1, 1, 0, 1, 0, 1])

    def test_get_match_array_signed_unsigned_and_binary(self):
        pass

    @unittest.skip("Not currently working.")
    def test_add_forbidden(self):
        m = numpy.ones([2, 3])
        preprocessors_list = ["pa", "pb"]
        classifier_list = ["ca", "cb", "cc"]
        cs = ConfigurationSpace()
        preprocessor = CategoricalHyperparameter(
            name="feature_preprocessor", choices=preprocessors_list
        )
        classifier = CategoricalHyperparameter(
            name="classifier", choices=classifier_list
        )
        cs.add_hyperparameter(preprocessor)
        cs.add_hyperparameter(classifier)
        new_cs = autosklearn.pipeline.create_searchspace_util.add_forbidden(
            conf_space=cs,
            node_0_list=preprocessors_list,
            node_1_list=classifier_list,
            matches=m,
            node_0_name="feature_preprocessor",
            node_1_name="classifier",
        )
        self.assertEqual(len(new_cs.forbidden_clauses), 0)
        self.assertIsInstance(new_cs, ConfigurationSpace)

        m[1, 1] = 0
        new_cs = autosklearn.pipeline.create_searchspace_util.add_forbidden(
            conf_space=cs,
            node_0_list=preprocessors_list,
            node_1_list=classifier_list,
            matches=m,
            node_0_name="feature_preprocessor",
            node_1_name="classifier",
        )
        self.assertEqual(len(new_cs.forbidden_clauses), 1)
        self.assertEqual(new_cs.forbidden_clauses[0].components[0].value, "cb")
        self.assertEqual(new_cs.forbidden_clauses[0].components[1].value, "pb")
        self.assertIsInstance(new_cs, ConfigurationSpace)



# -*- encoding: utf-8 -*-
import copy
import json
import logging.handlers
import multiprocessing
import os
import shutil
import sys
import tempfile

import numpy as np
from smac.tae import StatusType

from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.constants import (
    BINARY_CLASSIFICATION,
    MULTICLASS_CLASSIFICATION,
    MULTILABEL_CLASSIFICATION,
    REGRESSION,
)
from autosklearn.evaluation.test_evaluator import TestEvaluator, eval_t
from autosklearn.evaluation.util import read_queue
from autosklearn.metrics import accuracy, balanced_accuracy, f1_macro, r2
from autosklearn.util.pipeline import get_configuration_space

import unittest
import unittest.mock

import test.conftest

this_directory = os.path.dirname(__file__)
sys.path.append(this_directory)
from evaluation_util import (  # noqa (E402: module level import not at top of file)
    SCORER_LIST,
    BaseEvaluatorTest,
    get_dataset_getters,
    get_evaluation_backend,
    get_multiclass_classification_datamanager,
)

N_TEST_RUNS = 3


class Dummy(object):
    pass


class TestEvaluator_Test(BaseEvaluatorTest, unittest.TestCase):
    _multiprocess_can_split_ = True

    def test_datasets(self):
        for getter in get_dataset_getters():
            testname = "%s_%s" % (
                os.path.basename(__file__).replace(".pyc", "").replace(".py", ""),
                getter.__name__,
            )

            with self.subTest(testname):
                backend_mock = get_evaluation_backend()
                D = getter()
                D_ = copy.deepcopy(D)
                y = D.data["Y_train"]
                if len(y.shape) == 2 and y.shape[1] == 1:
                    D_.data["Y_train"] = y.flatten()
                backend_mock.load_datamanager.return_value = D_
                metric_lookup = {
                    MULTILABEL_CLASSIFICATION: f1_macro,
                    BINARY_CLASSIFICATION: accuracy,
                    MULTICLASS_CLASSIFICATION: accuracy,
                    REGRESSION: r2,
                }
                queue_ = multiprocessing.Queue()

                evaluator = TestEvaluator(
                    backend_mock,
                    queue_,
                    metrics=[metric_lookup[D.info["task"]]],
                    port=logging.handlers.DEFAULT_TCP_LOGGING_PORT,
                    additional_components=dict(),
                )

                evaluator.fit_predict_and_loss()
                return_value = read_queue(evaluator.queue)
                self.assertEqual(len(return_value), 1)
                self.assertEqual(len(return_value[0]), 3)
                self.assertTrue(np.isfinite(return_value[0]["loss"]))


class DummyDatamanager:
    def __init__(self):
        self.info = {"task": MULTICLASS_CLASSIFICATION, "is_sparse": False}
        self.feat_type = {
            0: "numerical",
            1: "Numerical",
            2: "numerical",
            3: "numerical",
        }


class FunctionsTest(unittest.TestCase):
    def setUp(self):
        self.queue = multiprocessing.Queue()
        self.configuration = get_configuration_space(
            DummyDatamanager()
        ).get_default_configuration()
        self.data = get_multiclass_classification_datamanager()
        self.tmp_dir = os.path.join(os.path.dirname(__file__), ".test_cv_functions")
        self.backend = unittest.mock.Mock(spec=Backend)
        self.backend.temporary_directory = tempfile.gettempdir()
        self.backend.load_datamanager.return_value = self.data
        self.dataset_name = json.dumps({"task_id": "test"})

        self.port = logging.handlers.DEFAULT_TCP_LOGGING_PORT

    def tearDown(self):
        try:
            shutil.rmtree(self.tmp_dir)
        except Exception:
            pass

    def test_eval_test(self):
        eval_t(
            queue=self.queue,
            backend=self.backend,
            config=self.configuration,
            metrics=[accuracy],
            seed=test.conftest.DEFAULT_SEED,
            num_run=1,
            scoring_functions=None,
            output_y_hat_optimization=False,
            include=None,
            exclude=None,
            disable_file_output=False,
            instance=self.dataset_name,
            port=self.port,
            additional_components=dict(),
        )
        return_value = read_queue(self.queue)
        self.assertEqual(len(return_value), 1)
        self.assertAlmostEqual(return_value[0]["loss"], 0.07999999999999996)
        self.assertEqual(return_value[0]["status"], StatusType.SUCCESS)
        self.assertNotIn("bac_metric", return_value[0]["additional_run_info"])

    def test_eval_test_multi_objective(self):
        metrics = {
            accuracy: 0.07999999999999996,
            balanced_accuracy: 0.05555555555555547,
        }
        eval_t(
            queue=self.queue,
            backend=self.backend,
            config=self.configuration,
            metrics=list(metrics.keys()),
            seed=test.conftest.DEFAULT_SEED,
            num_run=1,
            scoring_functions=None,
            output_y_hat_optimization=False,
            include=None,
            exclude=None,
            disable_file_output=False,
            instance=self.dataset_name,
            port=self.port,
            additional_components=dict(),
        )
        return_value = read_queue(self.queue)
        self.assertEqual(len(return_value), 1)
        for metric, loss in metrics.items():
            self.assertAlmostEqual(return_value[0]["loss"][metric.name], loss)
        self.assertEqual(return_value[0]["status"], StatusType.SUCCESS)
        self.assertNotIn("bac_metric", return_value[0]["additional_run_info"])

    def test_eval_test_all_loss_functions(self):
        eval_t(
            queue=self.queue,
            backend=self.backend,
            config=self.configuration,
            metrics=[accuracy],
            seed=1,
            num_run=1,
            scoring_functions=SCORER_LIST,
            output_y_hat_optimization=False,
            include=None,
            exclude=None,
            disable_file_output=False,
            instance=self.dataset_name,
            port=self.port,
            additional_components=dict(),
        )
        return_value = read_queue(self.queue)
        self.assertEqual(len(return_value), 1)

        # Note: All metric here should be minimized
        fixture = {
            "accuracy": 0.040000000000000036,
            "balanced_accuracy": 0.02777777777777779,
            "f1_macro": 0.0341005967604433,
            "f1_micro": 0.040000000000000036,
            "f1_weighted": 0.039693094629155934,
            "log_loss": 0.13966929787769913,
            "precision_macro": 0.03703703703703709,
            "precision_micro": 0.040000000000000036,
            "precision_weighted": 0.03555555555555556,
            "recall_macro": 0.02777777777777779,
            "recall_micro": 0.040000000000000036,
            "recall_weighted": 0.040000000000000036,
            "num_run": -1,
        }

        additional_run_info = return_value[0]["additional_run_info"]
        for key, value in fixture.items():
            self.assertAlmostEqual(additional_run_info[key], fixture[key], msg=key)
        self.assertEqual(
            len(additional_run_info),
            len(fixture) + 1,
            msg=sorted(additional_run_info.items()),
        )
        self.assertIn("duration", additional_run_info)
        self.assertAlmostEqual(return_value[0]["loss"], 0.040000000000000036)
        self.assertEqual(return_value[0]["status"], StatusType.SUCCESS)



from typing import Optional

import ConfigSpace.configuration_space

import autosklearn.pipeline.base
import autosklearn.pipeline.components.base
import autosklearn.pipeline.components.classification as classification
import autosklearn.pipeline.components.feature_preprocessing as feature_preprocessing
from autosklearn.askl_typing import FEAT_TYPE_TYPE

import unittest
import unittest.mock


class BasePipelineMock(autosklearn.pipeline.base.BasePipeline):
    def __init__(self, feat_type: Optional[FEAT_TYPE_TYPE] = None):
        self.feat_type = feat_type


class BaseTest(unittest.TestCase):
    def test_get_hyperparameter_configuration_space_3choices(self):
        cs = ConfigSpace.configuration_space.ConfigurationSpace()
        dataset_properties = {"target_type": "classification"}
        exclude = {}
        include = {}
        pipeline = [
            (
                "p0",
                feature_preprocessing.FeaturePreprocessorChoice(dataset_properties),
            ),
            (
                "p1",
                feature_preprocessing.FeaturePreprocessorChoice(dataset_properties),
            ),
            (
                "c",
                classification.ClassifierChoice(dataset_properties),
            ),
        ]

        base = BasePipelineMock()
        cs = base._get_base_search_space(
            cs=cs,
            dataset_properties=dataset_properties,
            exclude=exclude,
            include=include,
            pipeline=pipeline,
        )

        self.assertEqual(len(cs.get_hyperparameter("p0:__choice__").choices), 13)
        self.assertEqual(len(cs.get_hyperparameter("p1:__choice__").choices), 15)

        # for clause in sorted([str(clause) for clause in cs.forbidden_clauses]):
        #     print(clause)
        self.assertEqual(148, len(cs.forbidden_clauses))

        cs = ConfigSpace.configuration_space.ConfigurationSpace()
        dataset_properties = {"target_type": "classification", "signed": True}
        include = {"c": ["multinomial_nb"]}
        cs = base._get_base_search_space(
            cs=cs,
            dataset_properties=dataset_properties,
            exclude=exclude,
            include=include,
            pipeline=pipeline,
        )
        self.assertEqual(len(cs.get_hyperparameter("p0:__choice__").choices), 13)
        self.assertEqual(len(cs.get_hyperparameter("p1:__choice__").choices), 10)
        self.assertEqual(len(cs.get_hyperparameter("c:__choice__").choices), 1)
        # Mostly combinations of p0 making the data unsigned and p1 not
        # changing the values of the data points
        # for clause in sorted([str(clause) for clause in cs.forbidden_clauses]):
        #    print(clause)
        self.assertEqual(64, len(cs.forbidden_clauses))

        cs = ConfigSpace.configuration_space.ConfigurationSpace()
        dataset_properties = {"target_type": "classification", "signed": True}
        include = {}
        cs = base._get_base_search_space(
            cs=cs,
            dataset_properties=dataset_properties,
            exclude=exclude,
            include=include,
            pipeline=pipeline,
        )
        self.assertEqual(len(cs.get_hyperparameter("p0:__choice__").choices), 13)
        self.assertEqual(len(cs.get_hyperparameter("p1:__choice__").choices), 15)
        self.assertEqual(len(cs.get_hyperparameter("c:__choice__").choices), 16)
        # for clause in sorted([str(clause) for clause in cs.forbidden_clauses]):
        #    print(clause)
        self.assertEqual(110, len(cs.forbidden_clauses))

        cs = ConfigSpace.configuration_space.ConfigurationSpace()
        dataset_properties = {"target_type": "classification", "sparse": True}
        cs = base._get_base_search_space(
            cs=cs,
            dataset_properties=dataset_properties,
            exclude=exclude,
            include=include,
            pipeline=pipeline,
        )
        self.assertEqual(len(cs.get_hyperparameter("p0:__choice__").choices), 12)
        self.assertEqual(len(cs.get_hyperparameter("p1:__choice__").choices), 15)
        # for clause in sorted([str(clause) for clause in cs.forbidden_clauses]):
        #    print(clause)
        self.assertEqual(419, len(cs.forbidden_clauses))

        cs = ConfigSpace.configuration_space.ConfigurationSpace()
        dataset_properties = {
            "target_type": "classification",
            "sparse": True,
            "signed": True,
        }
        cs = base._get_base_search_space(
            cs=cs,
            dataset_properties=dataset_properties,
            exclude=exclude,
            include=include,
            pipeline=pipeline,
        )

        self.assertEqual(len(cs.get_hyperparameter("p0:__choice__").choices), 12)
        self.assertEqual(len(cs.get_hyperparameter("p1:__choice__").choices), 15)
        # Data is guaranteed to be positive in cases like densifier,
        # extra_trees_preproc, multinomial_nb -> less constraints
        # for clause in sorted([str(clause) for clause in cs.forbidden_clauses]):
        #    print(clause)
        self.assertEqual(359, len(cs.forbidden_clauses))

    def test_init_params_handling(self):
        """
        Makes sure that init params is properly passed to nodes

        Also, makes sure that _check_init_params_honored raises the expected exceptions
        """
        cs = ConfigSpace.configuration_space.ConfigurationSpace()
        base = BasePipelineMock()
        base.dataset_properties = {}

        # Make sure that component irrespective, we check the init params
        for node_type in [
            autosklearn.pipeline.components.base.AutoSklearnComponent,
            autosklearn.pipeline.components.base.AutoSklearnChoice,
            autosklearn.pipeline.base.BasePipeline,
        ]:

            # We have couple of posibilities
            for init_params, expected_init_params in [
                ({}, {}),
                (None, None),
                ({"M:key": "value"}, {"key": "value"}),
            ]:
                node = unittest.mock.Mock(
                    spec=autosklearn.pipeline.components.base.AutoSklearnComponent
                )
                node.get_hyperparameter_search_space.return_value = cs
                node.key = "value"
                base.steps = [("M", node)]
                base.set_hyperparameters(
                    cs.sample_configuration(), init_params=init_params
                )
                self.assertEqual(
                    node.set_hyperparameters.call_args[1]["init_params"],
                    expected_init_params,
                )

            # Check for proper exception raising
            node = unittest.mock.Mock(
                spec=autosklearn.pipeline.components.base.AutoSklearnComponent
            )
            node.get_hyperparameter_search_space.return_value = cs
            base.steps = [("M", node)]
            with self.assertRaisesRegex(
                ValueError, "Unsupported argument to init_params"
            ):
                base.set_hyperparameters(
                    cs.sample_configuration(), init_params={"key": "value"}
                )

            # An invalid node name is passed
            with self.assertRaisesRegex(
                ValueError, "The current node name specified via key"
            ):
                base.set_hyperparameters(
                    cs.sample_configuration(), init_params={"N:key": "value"}
                )

            # The value was not properly set -- Here it happens because the
            # object is a magic mock, calling the method doesn't set a new parameter
            with self.assertRaisesRegex(ValueError, "Cannot properly set the pair"):
                base.set_hyperparameters(
                    cs.sample_configuration(), init_params={"M:key": "value"}
                )

    def test_include_exclude_validation(self):
        """Makes sure include and exclude arguments are validated and raises
        expected exception on error
        """
        base = BasePipelineMock()
        dataset_properties = {"target_type": "classification"}
        base.dataset_properties = dataset_properties
        base.steps = [
            (
                "p0",
                feature_preprocessing.FeaturePreprocessorChoice(dataset_properties),
            ),
            (
                "p1",
                feature_preprocessing.FeaturePreprocessorChoice(dataset_properties),
            ),
            (
                "c",
                classification.ClassifierChoice(dataset_properties),
            ),
        ]

        def assert_value_error(include=None, exclude=None):
            base.include = include
            base.exclude = exclude
            with self.assertRaises(ValueError):
                base._validate_include_exclude_params()

        # Same key in include and exclude argument
        assert_value_error(include={"c": ["adaboost"]}, exclude={"c": ["sgd"]})

        # Invalid key in the exclude argument
        assert_value_error(exclude={"p2": ["pca"]})

        # Invalid value type for the key in the include argument
        assert_value_error(include={"c": ("adaboost", "sgd")}, exclude=None)

        # Empty list of the key in the include argument
        assert_value_error(include={"c": []})

        # Invalid component in the list value for the key in the include argument
        assert_value_error(include={"c": ["pca"]})

        # Case when all conditions passed for include and exclude
        base.include = {"c": ["adaboost", "sgd"]}
        base.exclude = {"p1": ["pca"]}
        self.assertIsNone(base._validate_include_exclude_params())



# -*- encoding: utf-8 -*-
from typing import Dict, List, Optional, Union

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace

from autosklearn.constants import (
    BINARY_CLASSIFICATION,
    MULTICLASS_CLASSIFICATION,
    MULTILABEL_CLASSIFICATION,
    MULTIOUTPUT_REGRESSION,
    REGRESSION_TASKS,
)
from autosklearn.data.abstract_data_manager import AbstractDataManager
from autosklearn.pipeline.classification import SimpleClassificationPipeline
from autosklearn.pipeline.regression import SimpleRegressionPipeline

__all__ = ["get_configuration_space"]


def get_configuration_space(
    datamanager: AbstractDataManager,
    include: Optional[Dict[str, List[str]]] = None,
    exclude: Optional[Dict[str, List[str]]] = None,
    random_state: Optional[Union[int, np.random.RandomState]] = None,
) -> ConfigurationSpace:
    """Get the configuration of a pipeline given some dataset info

    Parameters
    ----------
    datamanager: AbstractDataManager
        AbstractDataManager object storing all important information about the dataset

    include: Optional[Dict[str, List[str]]] = None
        A dictionary of what components to include for each pipeline step

    exclude: Optional[Dict[str, List[str]]] = None
        A dictionary of what components to exclude for each pipeline step

    random_state: Optional[Union[int, np.random.Randomstate]] = None
        The random state to use for seeding the ConfigSpace

    Returns
    -------
    ConfigurationSpace
        The configuration space for the pipeline
    """
    if datamanager.info["task"] in REGRESSION_TASKS:
        return _get_regression_configuration_space(
            datamanager, include, exclude, random_state
        )
    else:
        return _get_classification_configuration_space(
            datamanager, include, exclude, random_state
        )


def _get_regression_configuration_space(
    datamanager: AbstractDataManager,
    include: Optional[Dict[str, List[str]]],
    exclude: Optional[Dict[str, List[str]]],
    random_state: Optional[Union[int, np.random.RandomState]] = None,
) -> ConfigurationSpace:
    """Get the configuration of a regression pipeline given some dataset info

    Parameters
    ----------
    datamanager: AbstractDataManager
        AbstractDataManager object storing all important information about the dataset

    include: Optional[Dict[str, List[str]]] = None
        A dictionary of what components to include for each pipeline step

    exclude: Optional[Dict[str, List[str]]] = None
        A dictionary of what components to exclude for each pipeline step

    random_state: Optional[Union[int, np.random.Randomstate]] = None
        The random state to use for seeding the ConfigSpace

    Returns
    -------
    ConfigurationSpace
        The configuration space for the regression pipeline
    """
    task_type = datamanager.info["task"]
    sparse = False
    multioutput = False
    if task_type == MULTIOUTPUT_REGRESSION:
        multioutput = True

    if datamanager.info["is_sparse"] == 1:
        sparse = True

    dataset_properties = {"multioutput": multioutput, "sparse": sparse}

    configuration_space = SimpleRegressionPipeline(
        feat_type=datamanager.feat_type,
        dataset_properties=dataset_properties,
        include=include,
        exclude=exclude,
        random_state=random_state,
    ).get_hyperparameter_search_space(feat_type=datamanager.feat_type)
    return configuration_space


def _get_classification_configuration_space(
    datamanager: AbstractDataManager,
    include: Optional[Dict[str, List[str]]],
    exclude: Optional[Dict[str, List[str]]],
    random_state: Optional[Union[int, np.random.RandomState]] = None,
) -> ConfigurationSpace:
    """Get the configuration of a classification pipeline given some dataset info

    Parameters
    ----------
    datamanager: AbstractDataManager
         AbstractDataManager object storing all important information about the dataset

    include: Optional[Dict[str, List[str]]] = None
        A dictionary of what components to include for each pipeline step

    exclude: Optional[Dict[str, List[str]]] = None
        A dictionary of what components to exclude for each pipeline step

    random_state: Optional[Union[int, np.random.Randomstate]] = None
        The random state to use for seeding the ConfigSpace

    Returns
    -------
    ConfigurationSpace
        The configuration space for the classification pipeline
    """
    task_type = datamanager.info["task"]

    multilabel = False
    multiclass = False
    sparse = False

    if task_type == MULTILABEL_CLASSIFICATION:
        multilabel = True
    if task_type == MULTICLASS_CLASSIFICATION:
        multiclass = True
    if task_type == BINARY_CLASSIFICATION:
        pass

    if datamanager.info["is_sparse"] == 1:
        sparse = True

    dataset_properties = {
        "multilabel": multilabel,
        "multiclass": multiclass,
        "sparse": sparse,
    }

    return SimpleClassificationPipeline(
        feat_type=datamanager.feat_type,
        dataset_properties=dataset_properties,
        include=include,
        exclude=exclude,
        random_state=random_state,
    ).get_hyperparameter_search_space(feat_type=datamanager.feat_type)
